#!/usr/bin/env python

"""
Custom toyplot Mark to create SVG pie charts on nodes, and a
user-facing annotation function for plotting pie charts from
... data.

A standard CSS way to do piecharts is with many circle elements
piled up and using stroke-dasharray and stroke-offset. Its an
ugly hack, we're not using it.

An alternative is to draw filled path arcs in SVG. This is not so
different from what we already do to draw the tree edges, so I
preferred to use this.

<g class='pie-0'b style='stroke: black; stroke-width: 1'>
    <path d="M 1 0 A 1 1 0 0 1 0.8 0.5 L 0 0" style='fill:red'><path>
    <path d="M 1 0 A 0 1 0 0 1 0.8 0.5 L 0 0" style='fill:pink'><path>
</g>           |    |                  |
               move to start position on arc.
                    |                  |
                    arc: rx ry x-axis-rotation large-arc-floag sweep-flag x y
                         radius    always 0        %>50 or no    always 1  end
                                       |
                                       move to center of circle.
"""

import functools
import xml.etree.ElementTree as xml
from multipledispatch import dispatch
import toyplot
import numpy as np
import toyplot.html
from toyplot.coordinates import Cartesian
from toyplot.mark import Mark

from toytree.color import ToyColor
from toytree.color.src.concat import concat_style_fix_color
from toytree.drawing.src.mark_pie import PieChartMark


# Register multipledispatch to use the toyplot.html namespace
dispatch = functools.partial(dispatch, namespace=toyplot.html._namespace)


#######################################################################
@dispatch(Cartesian, PieChartMark, toyplot.html.RenderContext)
def _render(axes, mark, context):
    render_pie_chart(axes, mark, context)
#######################################################################


def render_pie_chart(axes: Cartesian, mark: Mark, context: toyplot.html.RenderContext) -> None:
    """Multidispatch registered render function for PieChartMarks.

    The PieChartMark object is a custom Mark and contains the data,
    coordinates, and styles. The colors are checked here during
    rendering.
    """
    # project coordinates to axes
    nodes_x = axes.project('x', mark.coordinates[:, 0])
    nodes_y = axes.project('y', mark.coordinates[:, 1])

    # create a group for pie node markers
    mark_xml = xml.SubElement(
        context.parent, "g",
        id=context.get_id(mark),
        attrib={"class": "toytree-mark-PieCharts"},
    )

    # fill dict w/ dicts {idx: {fill: ..., fill-opacity: ...}, ...}
    colors = {}
    for cidx, color in enumerate(mark.colors):
        colors[cidx] = {'fill': ToyColor(color)}

        # get shared inner stroke styles
        shared_style = {
            "stroke-linecap": "round",
            "stroke-width": mark.istroke_width,
        }
        shared_style.update({"stroke": ToyColor(mark.istroke).css})

    # render the pies as a group of path elements.
    for nidx in range(mark.coordinates.shape[0]):
        group = xml.SubElement(
            mark_xml, "g",
            attrib={'id': f'pie-{nidx}'},
            style=concat_style_fix_color(shared_style),
        )
        xpos = nodes_x[nidx] + mark.xshift
        ypos = nodes_y[nidx] + mark.yshift
        transform = (
            f"translate({xpos:.4f},{ypos:.4f}) "
            f"rotate({mark.rotate})"
        )
        group.set("transform", transform)

        # iterate over slices: e.g., [0.3, 0.5, 0.2]
        # sums = [0, 0.3, 0.8, 1.0]
        for cidx in range(mark.data[0].size):
            start_sum = mark.data[nidx][:cidx].sum()
            end_sum = mark.data[nidx][:cidx + 1].sum()

            # only set radius on circle since placement uses transform
            path = _get_pie_path(
                percent_start=start_sum,
                percent_end=end_sum,
                radius=mark.sizes[nidx],
            )
            xml.SubElement(
                group, "path",
                d=path,
                style=concat_style_fix_color(colors[cidx])
            )

        # add a circle to outline the node and
        # TODO: provide optional title hover
        ostyle = {
            "fill": (0, 0, 0, 0),
            "stroke-width": mark.ostroke_width,
        }
        ostyle.update({"stroke": ToyColor(mark.ostroke)})
        xml.SubElement(
            group, "circle",
            r=str(mark.sizes[nidx]),
            style=concat_style_fix_color(ostyle),
        )


def _get_pie_path(percent_start: float, percent_end: float, radius: float) -> str:
    """Return a SVG path for a circle arc.

    <path d='M end_x end_y A r r 0 flag 1 end_x end_y L 0 0'></path>
    """
    start = _get_radial_coordinates_for_percents(percent_start)
    end = _get_radial_coordinates_for_percents(percent_end)
    flag = int((percent_end - percent_start) > 0.5)
    return (
        f"M 0 0 L {radius * start[0]} {radius * start[1]} "
        f"A {radius} {radius} 0 {flag} 1 {radius * end[0]} {radius * end[1]} "
        f"L 0 0"
    )


def _get_radial_coordinates_for_percents(percent):
    """Return the coordinates on a circle of a percentage."""
    xpos = round(np.cos(2 * np.pi * percent), 7)
    ypos = round(np.sin(2 * np.pi * percent), 7)
    return xpos, ypos


if __name__ == "__main__":

    import toytree
    import toyplot

    TREE = toytree.rtree.bdtree(30, seed=123)
    c, a, m = TREE.draw(width=400, height=600, node_sizes=5)
    DATA = np.array([[0.5, 0.3, 0.2]] * (TREE.nnodes))
    COLORS = toytree.color.COLORS1
    TREE.annotate.add_node_pie_charts(axes=a, data=DATA, colors="Greys", mask=False)

    import toyplot.browser
    toyplot.browser.show(c)



#!/usr/bin/env python

"""Tree distance measures based on tree moves (SPR and NNI).

TODO.

References
----------
- 
"""


def tree_dist_nni():
    """

    Use the approach of Li et al. (1996) to approximate the Nearest 
    Neighbour Interchange distance (Robinson, 1971) between 
    phylogenetic trees.

    FROM TREEDIST:
    https://cran.r-project.org/web/packages/TreeDist/TreeDist.pdf

    In brief, this approximation algorithm works by identifying edges in one tree that do not match
    edges in the second. Each of these edges must undergo at least one NNI operation in order to
    reconcile the trees. Edges that match in both trees need never undergo an NNI operation, and divide
    each tree into smaller regions. By ’cutting’ matched edges into two, a tree can be divided into a
    number of regions that solely comprise unmatched edges.
    These regions can be viewed as separate trees that need to be reconciled. One way to reconcile
    these trees is to conduct a series of NNI operations that reduce a tree to a pectinate (caterpillar) tree,
    then to conduct an analogue of the mergesort algorithm. This takes at most n log n + O(n) NNI
    operations, and provides a loose upper bound on the NNI score. The maximum number of moves
    for an n-leaf tree (OEIS A182136) can be calculated exactly for small trees (Fack et al. 2002); this
    provides a tighter upper bound, but is unavailable for n > 12. NNIDiameter() reports the limits on
    this bound.    

    References
    ----------
    - Li et al. (1996)
    - Fack V, Lievens S, Van der Jeugt J (2002). “On the diameter of 
    the rotation graph of binary coupling trees.” Discrete Mathematics,
    245(1-3), 1–18.
    - Robinson DF (1971). “Comparison of labeled trees with valency 
    three.” Journal of Combinatorial Theory, Series B, 11(2), 105–119.
    """


def tree_dist_spr(tree1, tree2, max_dist: int=10) -> float:
    """Heuristic search algorithm to count N SPR moves separating two trees.
    
    Starting from tree1 all SPR moves that decrease the distance 
    ... continue until dist=0 or max_dist is exceeded.
    """


if __name__ == "__main__":
    pass



#!/usr/bin/env python

"""Phylogenetic comparative methods (pcm) subpackage.

All `pcm` source code is located in `toytree.pcm.src`, with user-facing 
functions exposed in the `api` and `tree_api` submodules. In this 
way, all functions that a user is meant to interact with can be 
accessed from top-level locations at either :mod:`toytree.pcm` 
(package-level usage) or the `.pcm` attribute of :class:`ToyTree`
instances (instance-level usage).

Package level API usage
>>> tree = toytree.rtree.unittree(ntips=10, treeheight=100, seed=123)   
>>> toytree.pcm.simulate_discrete_markov_data(tree, 3, "ER")

Instance level API usage
>>> tree = toytree.rtree.unittree(ntips=10, treeheight=100, seed=123)
>>> tree.pcm.simulate_discrete_markov_data(nstates=3, "ER")

Note
----
This package may be split into more subdivided packages in the future,
as it grows, such as `toytree.pcm.markov`, `toytree.pcm.phylocom`.
"""

# import the tree-api: tree.pcm.functions
# from .src.api_tree import PhyloCompAPI

# import the package-api functions: tree.pcm.functions
from toytree.pcm.src.vcv import (
    get_vcv_matrix_from_tree,
    get_corr_matrix_from_tree,
    # get_tree_from_vcv,
)
from toytree.pcm.src.traits import *
# from toytree.pcm.src.diversification import *
# from toytree.pcm.src.traits import *
# from toytree.pcm.src.phylocom import simulate_community_data



#!/usr/bin/env python

"""Tree moves used in heuristic tree search algorithms.

Tree 'moves' (perturbations) are used to search tree space and
usually applied with an optimality criterion to find a best scoring
tree under either parsimony or maximum likelihood hill-climbing
algorithms.

TODO
----
could implement using only .mod topo funcs on ToyTrees?

UNDER DEVELOPMENT
-----------------
- rooted tree moves need to iterate over placements of the root?

See also eSPR and eTBR moves used in Bayesian phylogenetics. These
modify the probability of proposed pruning points by updating a
'distance' parameter as part of the MCMC process. For example in
mrbayes. Not implemented here.
"""

from typing import Optional, TypeVar, Iterator
from loguru import logger
import numpy as np
import toytree
from toytree.utils import ToytreeError


logger = logger.bind(name="toytree")
ToyTree = TypeVar("ToyTree")


# FIXME: needs further work.
def move_spr_iter(tree: ToyTree, highlight: bool = False) -> Iterator[ToyTree]:
    """Generator to yield all trees within 1 SPR move of input tree.

    Returns a generator function to iterate through each of the
    unrooted trees that are within 1 subprune and regraft (SPR)
    moves from the input tree. This tree move operation is used to
    heuristically search tree space, and can find relatively large
    changes from the current tree compared to NNI.

    TODO
    ----
    Return only unique topologies, e.g., subtree X inserted to edge
    Y may not be unique from subtree Y inserted to edge X.
    """
    tree = tree.unroot()

    # iterate over internal edges, skip tips, root, and one root child
    # to select each subtree that could be extacted.
    for nidx in range(tree.ntips, tree.nnodes - 1):

        # get list of Nodes (edges) where subtree can be inserted. This
        # cannot be root, or a desc on the subtree Node, or the subtree itself.
        subtree = tree[nidx]
        edges = (
            set(range(tree.nnodes))
            - set((i._idx for i in subtree.iter_descendants()))
            - set((i._idx for i in subtree.iter_sisters()))
            - set((subtree._up._idx, ))
            - set((subtree._idx, ))
        )

        # iterate over each possible insertion point of this edge
        for iedge in edges:
            logger.info(f"{subtree} -> {tree[iedge]}")

            # create copy of unrooted starting tree
            ntree = tree.copy()
            new_sister = ntree[iedge]
            subtree = ntree[nidx]

            # connect subtree to new sister by inserting a new Node
            new_node = toytree.Node("new")
            new_node_parent = new_sister._up
            old_node = subtree._up
            old_node_parent = old_node._up
            new_node._up = new_node_parent
            new_node._children = (subtree, new_sister)

            # disconnect 7 from 5
            old_node._remove_child(subtree)

            # connect 2 nodes on either side of 7
            if old_node_parent:
                old_node_parent._remove_child(old_node)
                for child in old_node._children:
                    old_node_parent._add_child(child)
                    child._dist += old_node._dist
            else:
                children = sorted(old_node._children, key=lambda x: x.idx)[::-1]
                for child in children[1:]:
                    old_node._remove_child(child)
                    children[0]._add_child(child)
            del old_node

            # connect subtree to tree
            subtree._up = new_node
            new_sister._up = new_node
            if new_node_parent:
                new_node_parent._remove_child(new_sister)
                new_node_parent._add_child(new_node)

            # if old root is now a singleton b/c new_sister is one of its children
            if len(ntree.treenode.children) == 1:
                ntree.treenode = ntree.treenode.children[0]
                oldroot = ntree.treenode._up
                ntree.treenode._up = None
                del oldroot

            # if new_sister is now the root.
            elif new_sister == ntree.treenode:
                ntree.treenode = new_node
            ntree._update()

            # optionally add style highlights
            if highlight:
                ntree = style_tree(ntree)
                ntree.style.node_colors = 'white'
                ntree.style.edge_colors = ['black'] * ntree.nnodes
                ntree.style.edge_widths = [2] * ntree.nnodes

                # color edge green
                # ntree.style.edge_colors[new_node.idx] = toytree.color.COLORS2[0]
                # ntree.style.edge_widths[new_node.idx] = 5

                # # color edge orange
                # ntree.style.edge_colors[subtree.idx] = toytree.color.COLORS2[1]
                # ntree.style.edge_widths[subtree.idx] = 5

                # # color clade 1 orange
                # for edge in nchildren:
                #     ntree.style.edge_colors[edge.idx] = toytree.color.COLORS2[1]
                #     ntree.style.edge_widths[edge.idx] = 5

                # # color clade 1 purple
                # for edge in nsisters + (nparent,):
                #     ntree.style.edge_colors[edge.idx] = toytree.color.COLORS2[2]
                #     ntree.style.edge_widths[edge.idx] = 5
            yield ntree


def move_nni_iter(tree: ToyTree, highlight: bool = False) -> Iterator[ToyTree]:
    """Generator to yield all trees within one NNI of input tree.

    Returns a generator function to iterate through each of the
    `2 * (ntips - 3)` unrooted trees that are within one nearest-
    neighbor interchange (NNI) move of the input tree. This tree
    move operation is used to heuristically search tree space, and
    is best for finding small changes from the current tree.

    Parameters
    ----------
    tree: ToyTree
        Tree from which one NNI move will be performed.
    highlight: bool
        If True the returned tree's .style dict is modified to
        highlight the swapped edges when drawn w/o a treestyle.

    Example
    -------
    >>> tree = toytree.rtree.unittree(ntips=5, seed=123)
    >>> nni_gen = toytree.mod.move_nni_iter(tree, highlight=True)
    >>> for ntre in nni_gen:
    >>>     ntre.draw(layout='unrooted', use_edge_lengths=False)
    """
    tree = tree.unroot()

    # iterate over internal edges, skip tips, root, and one root child.
    for nidx in range(tree.ntips, tree.nnodes - 1):
        node = tree[nidx]

        # swap each child with one sister of selected Node
        children = node.children
        sister = node.get_sisters()[0]
        for pick in children:

            # make COPY of the tree
            ntree = tree.copy()

            # select Nodes from the COPIED tree.
            nnode = ntree[node._idx]
            nparent = ntree[node._up._idx]
            npick = ntree[pick._idx]
            nsister = ntree[sister._idx]
            nchildren = nnode.children
            nsisters = nnode.get_sisters()

            # collapse node (connect children to parent, remove self from parent)
            for child in nchildren:
                nparent._add_child(child)
            nparent._remove_child(nnode)

            # re-insert node
            new_node = toytree.Node()
            new_node.label = "new"
            nparent._add_child(new_node)
            nparent._remove_child(npick)
            new_node._add_child(npick)
            nparent._remove_child(nsister)
            new_node._add_child(nsister)

            # update Node idxs and coordinates
            # logger.info(f"NNI: {nnode} | {npick} <--> {nsister}")
            ntree._update()

            # optionally add style highlights
            if highlight:
                ntree = style_tree(ntree)
                ntree.style.node_colors = 'white'
                ntree.style.edge_colors = ['black'] * ntree.nnodes
                ntree.style.edge_widths = [2] * ntree.nnodes

                # color edge green
                ntree.style.edge_colors[new_node.idx] = toytree.color.COLORS2[0]
                ntree.style.edge_widths[new_node.idx] = 5

                # color clade 1 orange
                for edge in nchildren:
                    ntree.style.edge_colors[edge.idx] = toytree.color.COLORS2[1]
                    ntree.style.edge_widths[edge.idx] = 5

                # color clade 1 purple
                for edge in nsisters + (nparent,):
                    ntree.style.edge_colors[edge.idx] = toytree.color.COLORS2[2]
                    ntree.style.edge_widths[edge.idx] = 5
            yield ntree


def move_nni(
    tree: ToyTree,
    edge: Optional[int] = None,
    seed: Optional[int] = None,
    inplace: bool = False,
    highlight: bool = False,
) -> ToyTree:
    """Return a tree one nearest-neighbor-interchange from current tree.

    An edge is selected and a subtree from either side is randomly
    selected and swapped.

    Parameters
    ----------
    tree: ToyTree
        A tree on which to perform a single NNI move.
    edge: int or None
        The int idx label of an edge in the tree to perform NNI on,
        selected by the Node idx below the edge. If None then an
        internal edge will be randomly selected.
    seed: int or None
        Seed for the numpy random number generator.
    inplace: bool
        If True the tree is modified inplace rather than copied.
    highlight: bool
        If True the .style dict is modified to highlight changes.
    """
    # work with unrooted tree, optionally as a copy (much slower).
    # TODO: can this be done without requiring unrooting?
    if inplace:
        tree.unroot(inplace=True)
    else:
        tree = tree.unroot()

    # create the random generator
    rng = np.random.default_rng(seed)

    # select an internal edge, or tip, which selects parent internal
    internal_edges = range(tree.ntips, tree.nnodes - 1)
    if edge is None:
        edge = rng.choice(internal_edges)
    else:
        if edge not in internal_edges:
            raise ToytreeError(f"idx must be an internal edge: {internal_edges}")
    node = tree[edge]

    # get references to neighbor nodes
    parent = node.up
    children = node.children

    # select a child to swap with
    pick = rng.choice(children)
    sisters = node.get_sisters()
    sister = sisters[0]
    logger.info(f"NNI: edge={node}, swap={pick} <--> {sister}")

    # add labels when debugging
    # sister.label = "sister"
    # node.label = "node"
    # parent.label = "parent"
    # for idx, child in enumerate(children):
    #     child.label = f"child-{idx}"

    # collapse node (connect children to parent, remove self from parent)
    for child in children:
        parent._add_child(child)
    parent._remove_child(node)

    # re-insert node
    # TODO: can this be done re-using node? is it better for data?
    new_node = toytree.Node()
    new_node.label = "new"
    parent._add_child(new_node)
    parent._remove_child(pick)
    new_node._add_child(pick)
    parent._remove_child(sister)
    new_node._add_child(sister)

    # update Node idxs and coordinates
    tree._update()

    # optionally add style highlights
    if highlight:
        tree = style_tree(tree)
        tree.style.node_colors = 'white'
        tree.style.edge_colors = ['black'] * tree.nnodes
        tree.style.edge_widths = [2] * tree.nnodes

        # color edge green
        tree.style.edge_colors[new_node.idx] = toytree.color.COLORS2[0]
        tree.style.edge_widths[new_node.idx] = 5

        # color clade 1 orange
        for edg in children:
            tree.style.edge_colors[edg.idx] = toytree.color.COLORS2[1]
            tree.style.edge_widths[edg.idx] = 5

        # color clade 1 purple
        for edg in sisters + (parent,):
            tree.style.edge_colors[edg.idx] = toytree.color.COLORS2[2]
            tree.style.edge_widths[edg.idx] = 5
    return tree


# TODO: try to simplify more like in NNI
def move_spr(
    tree: ToyTree,
    seed: Optional[int] = None,
    inplace: bool = False,
    highlight: bool = False,
) -> ToyTree:
    """Return a rooted ToyTree one SPR move from the current tree.

    The returned tree will have a different topology from the starting
    tree, at an SPR distance of 1. It randomly samples a subtree to
    extract from the tree, and then reinserts the subtree at an edge
    that is not (1) one of its descendants; (2) its sister; (3) its
    parent; or (4) itself.

    Parameters
    ----------
    tree: ToyTree
        The tree to be modified by a tree move.
    seed: int or None
        Seed for numpy random number generator.
    inplace: bool
        If True the tree is modified in place, else a copy is returned.
    highlight: bool
        If True the .style dict of the returned ToyTree will be
        modified to show the edges that were involved in the tree
        move if drawn without a tree_style argument.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(ntips=8, seed=123)
    >>> new_tree = toytree.mod.move_nni(tree, highlight=True)
    >>> tree.draw();
    >>> new_tree.draw();
    """
    tree = tree if inplace else tree.copy()
    rng = np.random.default_rng(seed)

    # randomly select a subtree (any non-root Node)
    sidx = rng.choice(tree.nnodes - 1)
    subtree = tree[sidx]

    # get list of Nodes (edges) where subtree can be inserted. This
    # cannot be root, or a desc on the subtree Node, or the subtree itself.
    edges = (
        set(range(tree.nnodes)) -
        set((i._idx for i in subtree._iter_descendants())) -
        set((i._idx for i in subtree._iter_sisters())) -
        set((subtree._up._idx, )) -
        set((subtree._idx, ))
    )

    # sample an edge by its descendant Node
    new_sister = tree[rng.choice(list(edges))]
    # logger.info(f"SPR: {sidx} -> {new_sister.idx}; options={edges}")

    # connect subtree to new sister by inserting a new Node
    new_node = toytree.Node("new")
    new_node_parent = new_sister._up
    old_node = subtree._up
    old_node_parent = old_node._up
    new_node._up = new_node_parent
    new_node._children = (subtree, new_sister)

    # remove 12 and connect 11 to 14
    old_node._remove_child(subtree)
    if old_node_parent:
        old_node_parent._remove_child(old_node)
        for child in old_node._children:
            old_node_parent._add_child(child)
            child._dist += old_node._dist
    del old_node

    # connect subtree to tree
    subtree._up = new_node
    new_sister._up = new_node
    if new_node_parent:
        new_node_parent._remove_child(new_sister)
        new_node_parent._add_child(new_node)

    # if old root is now a singleton b/c new_sister is one of its children
    if len(tree.treenode.children) == 1:
        tree.treenode = tree.treenode.children[0]
        oldroot = tree.treenode._up
        tree.treenode._up = None
        del oldroot

    # if new_sister is now the root.
    elif new_sister == tree.treenode:
        tree.treenode = new_node
    tree._update()

    # optional: color edges of the subtree that was moved.
    if highlight:
        tree.style.edge_colors = ['black'] * tree.nnodes
        tree.style.node_colors = ['white'] * tree.nnodes
        tree.style.node_style.stroke_width = 1.5
        tree.style.node_sizes = 8
        tree.style.node_labels = "idx"
        tree.style.node_labels_style.font_size = 12
        tree.style.node_labels_style._toyplot_anchor_shift = -9
        tree.style.node_labels_style.baseline_shift = 7.5
        tree.style.use_edge_lengths = False

        # tree.get_mrca_node(*tips)
        for node in subtree._iter_descendants():
            tree.style.edge_colors[node.idx] = toytree.color.COLORS2[3]
            tree.style.node_colors[node.idx] = toytree.color.COLORS2[3]
        tree.style.node_colors[new_node.idx] = toytree.color.COLORS2[3]
    return tree


def highlight_edges(tree: ToyTree, *edges: toytree.Node) -> ToyTree:
    """Color edges"""
    tree.style.edge_colors = ['black'] * tree.nnodes
    tree.style.edge_widths = [2] * tree.nnodes
    tree.style.node_colors = 'white'
    for idx, edge in enumerate(edges):
        tree.style.edge_colors[edge.idx] = toytree.color.COLORS2[idx]
        tree.style.edge_widths[edge.idx] *= 2
    return tree


def style_tree(tree: ToyTree) -> ToyTree:
    """Add style to show parts of tree that moved.
    """
    tree.style.layout = "unroot"
    tree.style.use_edge_lengths = False
    tree.style.node_style.stroke_width = 1.5
    tree.style.node_sizes = 6
    tree.style.node_labels = "idx"
    tree.style.node_labels_style.font_size = 12
    tree.style.node_labels_style._toyplot_anchor_shift = 8
    tree.style.node_labels_style.baseline_shift = 8
    tree.style.tip_labels_style._toyplot_anchor_shift = -8
    tree.style.tip_labels_style.baseline_shift = 8
    tree.style.tip_labels_style.font_size = 14
    return tree


# def move_nni_search(
#     tree: ToyTree,
#     criterion: Callable,
#     tolerance: float,
#     ) -> Iterator[ToyTree]:
#     """Return an infinite generator of sequential NNI.

#     This function can be used to perform a greedy hill-climbing
#     algorithm. It finds all trees within one NNI move from the
#     current tree, and applies the criterion function to each. The
#     tree with the lowest score will be yielded, and becomes the
#     source from NNI moves in the next iteration. If no trees have
#     a greater score then the generator ends.

#     Note
#     ----
#     To find a best-scoring tree a search should be started from many
#     different random starting trees.

#     Examples
#     --------
#     >>> tree = toytree.rtree.unittree(10, seed=123)
#     >>> search = nni_search(tree)
#     >>> path = [i for i in search]
#     >>> path[-1].draw()
#     """
#     pass
#     # TODO
#     # score = 0
#     # while 1:
#     #     for idx, proposal in enumerate(move_iter_nni):
#     #         score = criterion(proposal)
#     #         scores.append(score)
#     #     yield proposal


# def move_nni(
#     tree: ToyTree,
#     idx1: Optional[int]=None,
#     idx2: Optional[int]=None,
#     seed: Optional[int]=None,
#     inplace: bool=False,
#     highlight: bool=False,
#     ) -> ToyTree:
#     """Return a rooted ToyTree one SPR move from the current tree.

#     The returned tree will have a different topology from the starting
#     tree, at an SPR distance of 1. It randomly samples a subtree to
#     extract from the tree, and then reinserts the subtree at an edge
#     that is not (1) one of its descendants; (2) its sister; (3) its
#     parent; or (4) itself.

#     Parameters
#     ----------
#     tree: ToyTree
#         The tree to be modified by a tree move.
#     idx1: int or None
#         Node index of the first clade to be swapped with another. If
#         None then it is randomly sampled.
#     idx2: int or None
#         Node index of the second clade to be swapped with another. If
#         None then it is randomly sampled given limitations of idx1.
#     seed: int or None
#         Seed for numpy random number generator.
#     inplace: bool
#         If True the tree is modified in place, else a copy is returned.
#     highlight: bool
#         If True the .style dict of the returned ToyTree will be
#         modified to show the edges that were involved in the tree
#         move if drawn without a tree_style argument.

#     Examples
#     --------
#     >>> tree = toytree.rtree.unittree(ntips=8, seed=123)
#     >>> new_tree = toytree.mod.move_nni(tree, highlight=True)
#     >>> tree.draw();
#     >>> new_tree.draw();
#     """
#     # work with unrooted tree, optionally as a copy (much slower).
#     if inplace:
#         tree.unroot(inplace=True)
#     else:
#         tree = tree.unroot()

#     # create the random generator
#     rng = np.random.default_rng(seed)

#     # randomly select a subtree (any non-root Node)
#     node1 = tree[rng.choice(tree.nnodes - 1)] if idx1 is None else tree[idx1]

#     # get list of Nodes (edges) that can be swapped with node1.
#     # 1. starts with all possible Nodes except root.
#     # 2. removes descendants of node1
#     # 3. removes sisters of node1 (uninformative interchange)
#     # 4. remove parent of node1
#     # 5. remove node1
#     edges = (
#         set(range(tree.nnodes - 1)) -
#         set((i._idx for i in node1._iter_descendants())) -
#         set((i._idx for i in node1._iter_sisters())) -
#         set((node1._up._idx, )) -
#         set((node1._idx, ))
#     )

#     # sample an edge by its descendant Node
#     node2 = tree[rng.choice(list(edges))] if idx2 is None else tree[idx2]

#     # raise error if user-entered pair is invalid
#     if node2.idx not in edges:
#         raise ToytreeError(
#             f"invalid NNI move: {node2} not within valid pairs from {node1}: {edges}")

#     # debugging
#     logger.info(f"NNI: {node1} -> {node2}; options={edges}")

#     # re-attach Nodes using references saved before re-attaching
#     parent_1 = node1._up._idx
#     parent_2 = node2._up._idx
#     tree[parent_1]._remove_child(node1)
#     tree[parent_2]._remove_child(node2)
#     tree[parent_1]._add_child(node2)
#     tree[parent_2]._add_child(node1)

#     # store idx of both parents.
#     parent_1 = node1._up._idx
#     parent_2 = node2._up._idx

#     # remove swapping subtrees from the tree
#     tree[parent_1]._remove_child(node1)
#     tree[parent_2]._remove_child(node2)

#     # add the subtrees (node1 to parent2 and node2 to parent1)
#     tree[parent_1]._add_child(node2)
#     tree[parent_2]._add_child(node1)

#     # call toytree update routine
#     tree._update()

#     # add style to show subtrees swapped
#     if highlight:
#         tree.style.edge_colors = ['black'] * tree.nnodes
#         tree.style.node_colors = ['white'] * tree.nnodes
#         tree.style.node_style.stroke_width = 1.5
#         tree.style.node_sizes = 15
#         tree.style.node_labels = "idx"
#         tree.style.node_labels_style.font_size = 8
#         tree.style.tip_labels_style._toyplot_anchor_shift = -12
#         tree.style.tip_labels_style.baseline_shift = 9

#         tree.style.node_colors[node1.idx] = toytree.color.COLORS2[0]
#         for node in node1._iter_descendants():
#             tree.style.edge_colors[node.idx] = toytree.color.COLORS2[0]
#             tree.style.node_colors[node.idx] = toytree.color.COLORS2[0]
#         tree.style.node_colors[node2.idx] = toytree.color.COLORS2[3]
#         for node in node2._iter_descendants():
#             tree.style.edge_colors[node.idx] = toytree.color.COLORS2[3]
#             tree.style.node_colors[node.idx] = toytree.color.COLORS2[3]
#     return tree



# def get_all_neighbors_nni(
#     tree: ToyTree,
#     node=None,
#     highlight=False,
#     seed=None,
#     quiet=True,
#     ):
#     """Return a list of all trees within 1 NNI move from input tree.

#     Given a tree performs multiple NNI to get all the neighbors
#     given a node. If not is not provided, it is selected randomly.

#     Parameters
#     ----------
#     ...
#     """
#     tree = tree.copy()
#     tree = tree.unroot()

#     if not quiet: print(f"There are {2 * (tree.ntips - 3)} expected nearest neighbors for the given tree")

#     rng = np.random.default_rng(seed)


#     # randomly select first subtree (any non-root Node)
#     if node == None:
#         f_idx = rng.choice(tree.nnodes - 1)
#     else:
#         f_idx = node #use node specified by user

#     subtree_a = tree[f_idx]


#     # Check available nodes to select second subtree
#     # It should follow the following statements
#     available_nodes = (
#         set(range(tree.nnodes - 1)) # set with all possible nodes but the root
#         - set((i._idx for i in subtree_a._iter_descendants())) # remove descendants of subtree
#         - set((i._idx for i in subtree_a._iter_sisters())) # remove sisters of subtree (to avoid uninformative interchange)
#         - set((subtree_a._up._idx, )) # remove  parental of subtree (to avoid pick a subtree with subtree_a in there)
#         - set((subtree_a._idx, )) # remove subtree node itself
#     )
#     neighbor_trees = []
#     for available_node in available_nodes:
#         neighbor_trees.append(one_nni(tree, force=(f_idx, available_node), highlight=highlight, quiet=True, seed=seed))

#     return neighbor_trees


if __name__ == "__main__":

    toytree.set_log_level("INFO")
    NTIPS = 5
    TREE = toytree.rtree.unittree(NTIPS, seed=333).unroot()

    # draw the original tree
    c0, _, _ = TREE.draw(
        layout='unroot',
        use_edge_lengths=False,
        # tip_labels_style={"baseline-shift": 15},
        node_labels="idx",
        node_sizes=13,
    )

    # get mtree with all trees in NNI generator
    GEN = move_nni_iter(TREE, highlight=True)
    # GEN = move_spr_iter(TREE, highlight=True)
    MTRE = toytree.mtree(list(GEN))

    # get shape of tree drawing grid
    SIZE = 2 * (NTIPS - 3)
    SHAPE = (int(SIZE / 2), 2)

    # draw all trees within one NNI move
    c1, _, _ = MTRE.draw(
        shape=SHAPE,
        layout='unroot',
        use_edge_lengths=False,
        tip_labels_style={"font-size": 14},
    )

    # show result in default browser
    toytree.utils.show([c0, c1], new=False)



#!/usr/bin/env python

"""Submodule for quartet distance functions.

Authors: Deren Eaton and Scarlet Ming-sha Au

Computing Tree Similarity Metrics based on Quartets

All Tree Similarity Metrics based on Quartets revolve
around the following terms:

Q = Total possible quartets
S = Resolved in the same way between the two trees
D = Resolved differently between the two trees
R1 = Unresolved in tree 1, resolved in tree 2
R2 = Unresolved in tree 2, resolved in tree1
U = unresolved in both trees
N = S + D + R1 + R2 + U

Note
----
distance = 1 - similarity

References
----------
- Estabrook GF, McMorris FR, Meacham CA (1985). “Comparison of undirected
  phylogenetic trees based on subtrees of four evolutionary units.”
  Systematic Zoology, 34(2), 193--200. doi:10.2307/2413326 .
- ...
"""

from typing import Mapping
import itertools
import pandas as pd
import numpy as np
from toytree.core import ToyTree
from toytree.core.apis import TreeDistanceAPI, add_subpackage_method
import toytree


__all__ = [
    "get_treedist_quartets",
]


def get_quartet_resolutions_table(tree: ToyTree, df: bool = False) -> np.ndarray:
    """Return an array of quartet resolutions.

    The returned table can be an array of dataframe and is ordered
    alphanumerically by the names of the tips in the tree. If the tree
    has duplicate tip names an error will be raised.
    """
    snames = sorted(tree.get_tip_labels())
    assert len(snames) == len(set(snames)), "duplicate tip names are not allowed"

    # an empty array to enumerate all quartets and their resolutions
    ntips = toytree.enum.get_n_quartets(tree.ntips)
    arr = np.zeros(ntips, dtype=np.int_)

    # a dict to map {tipset: resolved} as {frozenset(abcd): tuple(abcd)}
    # note: this has a large memory usage for large trees.
    riter = tree.enum.iter_quartets(collapse=True, type=tuple, sort=True)
    rdict = {frozenset(i): i for i in riter}

    # iterator over tuples of sorted tip names: abcd, abce, abde
    iter_snames = itertools.combinations(snames, 4)
    for idx, qrt in enumerate(iter_snames):
        resolved = rdict.get(frozenset(qrt))
        if resolved:
            arr[idx] = qrt.index(resolved[1])

    # optionally return as df with index labels for debugging
    if df:
        return pd.Series(arr, index=list(itertools.combinations(snames, 4)))
    return arr


def get_quartet_comparison(tree1: ToyTree, tree2: ToyTree) -> Mapping[str, int]:
    """Return dict of quartet similarity/resolution data for two trees.

    This is used internally. Users should call `get_quartet_metrics`.
    """
    # require trees to share the same tips
    assert set(tree1.get_tip_labels()) == set(tree2.get_tip_labels())

    # get each quartet resolution array
    arr1 = get_quartet_resolutions_table(tree1)
    arr2 = get_quartet_resolutions_table(tree2)

    # stats dict
    data = {}

    # mask to select only quartes resolved in both trees
    mask = (arr1 != 0) & (arr2 != 0)

    # compute Q = total possible number of quartets
    data["Q"] = arr1.shape[0]

    # S = total number of quartets resolved in the same way in both trees
    data["S"] = sum(arr1[mask] == arr2[mask])

    # D = symmetric diff of resolved quartets
    data["D"] = sum(arr1[mask] != arr2[mask])

    # U = number of quartets unresolved in both trees
    data["U"] = sum((arr1 == 0) & (arr2 == 0))

    # R1 = number of quartets resolved in tree1, but unresolved in tree2
    data["R1"] = sum((arr1 != 0) & (arr2 == 0))

    # R2 = number of quartets resolved in tree2, but unresolved in tree1
    data["R2"] = sum((arr1 == 0) & (arr2 != 0))

    # N = S + D + R1 + R2 + U
    data["N"] = data["S"] + data["D"] + data["R1"] + data["R2"] + data["U"]
    return data


######################################################################
# COMPUTE METRICS GIVEN A DATA DICT WITH KEYS D, S, R1, R2, U, N, Q
# calculated in `get_quartet_comparison()`
######################################################################


def get_qrt_metric_do_not_conflict(data: Mapping[str, int]) -> float:
    return (data["S"] + data["R1"] + data["R2"] + data["U"]) / data["N"]


def get_qrt_metric_explicitly_agree(data: Mapping[str, int]) -> float:
    return data["S"] / data["N"]


def get_qrt_metric_strict_joint_assertions(data: Mapping[str, int]) -> float:
    return data["S"] / (data["S"] + data["D"])


def get_qrt_metric_semistrict_joint_assertions(data: Mapping[str, int]) -> float:
    return data["S"] / (data["S"] + data["D"] + data["U"])


def get_qrt_metric_steel_and_penny(data: Mapping[str, int]) -> float:
    return (data["S"] + data["U"]) / data["N"]


def get_qrt_metric_symmetric_difference(data: Mapping[str, int]) -> float:
    part1 = 2 * data["D"] + data["R1"] + data["R2"]
    part2 = (2 * data["D"]) + (2 * data["S"]) + data["R1"] + data["R2"]
    return 1 - (part1 / part2)


def get_qrt_metric_symmetric_divergence(data: Mapping[str, int]) -> float:
    return (2 * data["D"] + data["R1"] + data["R2"]) / data["N"]


def get_qrt_metric_similarity_to_reference(data: Mapping[str, int]) -> float:
    part1 = data["S"] + ((data["R1"] + data["R2"] + data["U"]) / 3)
    return part1 / data["Q"]


def get_qrt_metric_marczewski_steinhaus(data: Mapping[str, int]) -> float:
    part1 = (2 * data["D"]) + data["R1"] + data["R2"]
    part2 = (2 * data["D"]) + data["S"] + data["R1"] + data["R2"]
    return 1 - part1 / part2


QUARTET_METRICS = {
    "do_not_conflict": get_qrt_metric_do_not_conflict,
    "explicitly_agree": get_qrt_metric_explicitly_agree,
    "strict_joint_assertions": get_qrt_metric_strict_joint_assertions,
    "semistrict_joint_assertions": get_qrt_metric_semistrict_joint_assertions,
    "steel_and_penny": get_qrt_metric_steel_and_penny,
    "symmetric_difference": get_qrt_metric_symmetric_difference,
    "symmetric_divergence": get_qrt_metric_symmetric_divergence,
    "similarity_to_reference": get_qrt_metric_similarity_to_reference,
    "marczewski_steinhaus": get_qrt_metric_marczewski_steinhaus,
}

######################################################################
######################################################################
######################################################################


def get_quartet_metric(
    tree1: ToyTree,
    tree2: ToyTree,
    metric: str = None,
    similarity: bool = False,
) -> float:
    """Return a single quartet metric for two trees.

    Metrics
    -------
    ...

    Parameters
    ----------
    tree1: ToyTree
        A tree.
    tree1: ToyTree
        A tree with the same tip labels as tree1.
    metric: str or None
        A quartet metric (see 'Metrics' above) to compute.
    similarity: bool
        True returns similarity score, False returns a tree distance
        metric (1-similarity). Default is False.

    Example
    -------
    >>> tree1 = toytree.rtree.unittree(8, seed=123)
    >>> tree2 = toytree.rtree.unittree(8, seed=321)
    >>> get_quartet_metric(tree1, tree2)
    >>> # {'Q': 70, 'S': 70, 'D': 0, 'U': 0, 'R1': 0, 'R2': 0, 'N': 140}
    >>> get_quartet_metric(tree1, tree2, "steel_and_penny")
    >>> # 0.5
    """
    data = get_quartet_comparison(tree1, tree2)
    if metric is None:
        return data
    if metric in QUARTET_METRICS:
        value = QUARTET_METRICS[metric](data)
        if not similarity:
            value = 1 - value
        return value
    raise ValueError(
        f"metric {metric} not recognized, must be one of "
        f"{sorted(QUARTET_METRICS.keys())}.")


@add_subpackage_method(TreeDistanceAPI)
def get_treedist_quartets(
    tree1: ToyTree,
    tree2: ToyTree,
    similarity: bool = False,
) -> pd.Series:
    """Return a pd.Series with all quartet metrics for two trees.

    This returns all quartet metrics computed between two trees, since
    once quartets are enumerated and compared calculating the metrics
    is fast and simple.

    Parameters
    ----------
    tree1: ToyTree
        A tree.
    tree1: ToyTree
        A tree with the same tip labels as tree1.
    similarity: bool
        True returns similarity score, False returns a tree distance
        metric (1-similarity). Default is False.

    Examples
    --------
    >>> tree1 = toytree.rtree.unittree(8, seed=123)
    >>> tree2 = toytree.rtree.unittree(8, seed=123)
    >>> tree2.mod.collapse_nodes(10, 11, inplace=True)
    >>> get_treedist_quartet_metrics(tree1, tree2, similarity=False)
    >>> # Q                              70.000000
    >>> # S                              42.000000
    >>> # D                               0.000000
    >>> # U                               0.000000
    >>> # R1                             28.000000
    >>> # R2                              0.000000
    >>> # N                              70.000000
    >>> # do_not_conflict                 0.000000
    >>> # explicitly_agree                0.400000
    >>> # strict_joint_assertions         0.000000
    >>> # semistrict_joint_assertions     0.000000
    >>> # steel_and_penny                 0.400000
    >>> # symmetric_difference            0.250000
    >>> # symmetric_divergence            0.600000
    >>> # similarity_to_reference         0.266667
    >>> # marczewski_steinhaus            0.400000

    Quartet Data
    -------------
    ------------------------------------------
    | Q    | All possible quartets           |
    | S    | Resolved same in both trees     |
    | D    | Resolved diff between trees     |
    | U    | Unresolved in both trees        |
    | R1   | Unresolved in tree1 not tree2   |
    | R2   | Unresolved in tree2 not tree1   |
    | N    | S + D + U + R1 + R2             |
    ------------------------------------------

    Quartet Similarity Metrics
    --------------------------
    | name                    | compute               | reference   |
    --------------------------------------------------------------------------
    | DC: Do Not Conflict     | (S + R1 + R2 + U) / N | Estabrook et al. (1985) |
    | EA: Explicity Agree     | S / N                 | Estabrook et al. (1985) |
    | SJA: Strict Joint Assert| S / (S + D)           | Estabrook et al. (1985) |
    | SSJA: Semi-Strict ''    | S / (S + D + U)       | Estabrook et al. (1985) |
    | dQ: Steel and Penny     | (S + U) / N           | Steel and Penny (1993)  |
    | SD: Symmetric diff      | (2d + r1 + r2) / (2d + 2s + r1 + r2) | Day (1986) |
    | MS: Marczewski-Steinhaus| (2d + r1 + r2) / (2d + s + r1 + r2)| Marczewski and Steinhaus (1958) |
    | SV: Symmetric Divergence| (2d + r1 + r2) / N   | Smith 2019 |
    | S2R: Similarity to Ref  | (s + (r1 + r2 + u) / 3) / Q | Asher and Smith (2022) |
    --------------------------------------------------------------------------

    Reference
    ---------
    https://ms609.github.io/Quartet/reference/SimilarityMetrics.html
    """
    data = get_quartet_comparison(tree1, tree2)
    methods = list(data.keys()) + list(QUARTET_METRICS.keys())
    frame = pd.Series(index=methods, dtype=np.float64)
    for key, value in data.items():
        frame[key] = value
    for metric, method in QUARTET_METRICS.items():
        value = method(data)
        if not similarity:
            value = 1 - value
        frame[metric] = value
    return frame


if __name__ == "__main__":

    TREE1 = toytree.rtree.unittree(6, seed=123)
    TREE2 = toytree.rtree.unittree(6, seed=321)
    TREE2 = TREE2.mod.collapse_nodes(8)

    treedf1 = get_quartet_resolutions_table(TREE1, df=True)
    treedf2 = get_quartet_resolutions_table(TREE2, df=True)
    print(treedf2)

    data = get_quartet_comparison(TREE1, TREE2)
    print(data)

    # ...
    tree1 = toytree.rtree.unittree(8, seed=123)
    tree2 = toytree.rtree.unittree(8, seed=321)
    print(get_quartet_metric(tree1, tree2))
    print(get_quartet_metric(tree1, tree2, "steel_and_penny"))

    tree1 = toytree.rtree.unittree(8, seed=123)
    tree2 = toytree.rtree.unittree(8, seed=1233)
    tree2.mod.collapse_nodes(10, 11, inplace=True)
    print(get_treedist_quartets(tree1, tree2, similarity=False))



#!/usr/bin/env python

"""Generators to sample quartets of tips from a tree.

The primary function `iter_quartets` can be used as a generator to
yield quartet subtrees from a larger tree. This function is quite
fast and includes options for sorting the output, and transforming
its format to return Node objects, names, or any arbitrary feature
of Nodes. See examples.

Methods
-------
Get fast unordered sets of all combinations of 4 tip Nodes in a tree
>>> tree.enum._iter_unresolved_quartet_sets()   # {0, 1, 2, 3}, ...

Get name-ordered tuples of Nodes for each quartet induced by bipartitions in a tree.
>>> tree.enum.iter_quartets()                   # ((0, 1), (2, 3)), ...

See Also
--------
Get number of quartets induced by the splits in a tree.
>>> tree.enum.get_n_quartets()                  # 5

Format
------
Quartets represent a sample from a bipartition or quadripartition
where there is a split, e.g. `AB|CD`, separating to sets of items.
The order of the items within each partition of the quartet is not
often of interest, but it is nice to have a consistent sort option in
case it is useful.

Supported:
- ({'a', 'b'}, {'c', 'd'})  # type=set, collapse=False; sort affects order of p1,p2
- (('a', 'b'), ('c', 'd'))  # type=tuple, collapse=False; sort affects order of p1,p2 and within each p
- ('a', 'b', 'c', 'd')      # type=tuple, collapse=True; same as above, imagine middle split is still there.

Not supported:
- ({'a', 'b', 'c', 'd'})    # type=set, collapse=True; split info lost.
"""

from typing import TypeVar, Iterator, Tuple, Optional, Set, Callable
import itertools
from loguru import logger
from toytree import Node, ToyTree
from toytree.core.apis import TreeEnumAPI, add_subpackage_method, add_toytree_method

logger = logger.bind(name="toytree")
Query = TypeVar("Query")


__all__ = [
    "_iter_unresolved_quartet_sets",
    "_iter_quartet_sets",
    "iter_quartets",
]


def _iter_unresolved_quartet_sets(tree: ToyTree, feature: str = None) -> Iterator[Set]:
    """Generator to yield all combinations of four tip Nodes.

    This is simpler and faster than toytree.enum.iter_quartets because
    it does not find the resolution of quartets, but rather only finds
    all possible combinations. Note the intentional use of a set to
    indicate they are unordered.
    """
    # select function to format returned quartet set
    if feature is None:
        sformat = set
    else:
        def sformat(qrt) -> Set[str]:
            return set(getattr(node, feature) for node in qrt)
    for qrt in itertools.combinations(tree[:tree.ntips], 4):
        yield sformat(qrt)


@add_subpackage_method(TreeEnumAPI)
def _iter_quartet_sets(
    tree: ToyTree,
    feature: Optional[str] = None,
    quadripartitions: bool = False,
) -> Iterator[Tuple[Node, Node, Node, Node]]:
    """Generator to yield quartets induced by edges in a tree.

    This yields all quartets (4-sample subtrees) that exist within
    a larger tree. The set of possible quartets is not affected by
    tree rooting, but is affected by collapsed edges (polytomies),
    which reduce the number of quartets.

    Quartets are returned as Tuple[Node, Node, Node, Node], or Tuple
    of the requested features of Nodes, where e.g. ('a', 'b', 'c', 'd')
    implies the quartet `ab|cd`. The order in which quartets are
    yielded depends on the topology and rooting. The order of yielded
    quartets is in Node idx traversal order, where the first two Nodes
    are below the edge, and the second two above.

    Parameters
    ----------
    feature: str
        Feature used to represent Nodes on either side of bipartitions.
        Default is "name". None will return Node objects. Other Node
        features can be used but be aware if using quartets to compare
        among trees that 'idx' changes depending on topology, and other
        features may not be unique among Nodes.
    quadripartitions: bool
        If True then quartets are only returned that are induced by
        quadripartitite splits in a the tree. This is a subset of the
        quartets induced by bipartitions, since the tip Nodes must come
        from four different clades from each edge/split.

    Example
    -------
    >>> tree = toytree.rtree.unittree(5, seed=123)

    >>> sorted(tree.iter_quartets())
    >>> # (('r0', 'r1'), ('r2', 'r3'))
    >>> # (('r0', 'r1'), ('r2', 'r4'))
    >>> # (('r0', 'r1'), ('r3', 'r4'))
    >>> # (('r0', 'r2'), ('r3', 'r4'))
    >>> # (('r1', 'r2'), ('r3', 'r4'))

    """
    # find quartets induced by splits in the tree. This is necessary
    # compared to a simple call of `itertools.combinations(names, 4)`
    # because we indicate which pair of nodes is on which side of each
    # split. We also reduce number of quartets if any polytomies exist.

    # default args for iter_[bipart,qparts] to get Tuple[Node, ...]
    kwargs = dict(feature=feature, type=set, sort=False)

    # generate subset of quartets from quadripartitions
    if quadripartitions:
        for (i, j), (x, y) in tree.enum.iter_quadripartitions(**kwargs):
            pairgen = itertools.product(i, j, x, y)
            for i, j, x, y in pairgen:
                yield i, j, x, y

    # generate full collection of quartets from bipartitions
    else:
        observed = set()
        for below, above in tree.iter_bipartitions(**kwargs):
            # generator to sample 2 from either side of each bipart
            pairgen = itertools.product(
                itertools.combinations(below, 2),
                itertools.combinations(above, 2),
            )
            for (i, j), (x, y) in pairgen:
                qrt = frozenset((i, j, x, y))
                if qrt not in observed:
                    yield i, j, x, y
                    observed.add(qrt)


@add_toytree_method(ToyTree)
@add_subpackage_method(TreeEnumAPI)
def iter_quartets(
    tree: ToyTree,
    feature: Optional[str] = 'name',
    type: Callable = set,
    sort: bool = False,
    collapse: bool = False,
    quadripartitions: bool = False,
) -> Iterator:
    """Generator to yield quartets induced by edges in a tree.

    This yields all quartets (4-sample subtrees) that exist within
    a larger tree. The set of possible quartets is not affected by
    tree rooting, but is affected by collapsed edges (polytomies),
    which reduce the number of quartets.

    Quartets are returned as Tuple[Node, Node, Node, Node], or Tuple
    of the requested features of Nodes, where e.g. ('a', 'b', 'c', 'd')
    implies the quartet `ab|cd`. The order in which quartets are
    yielded depends on the topology and rooting, and is in Node idx
    traversal order, where the first two Nodes are below the edge, and
    the second two above. This can be changed to a consistent name
    sorted order for each split partition using `sort=True`.

    Parameters
    ----------
    feature: str
        Feature used to represent Nodes on either side of bipartitions.
        Default is "name". None will return Node objects. Other Node
        features can be used but be aware if using quartets to compare
        among trees that 'idx' changes depending on topology, and other
        features may not be unique among Nodes.
    type: Callable
        The type of collection used to represent a partition. Default
        is `set` to return a tuple of sets, but another useful option
        is `tuple`, which returns a tuple of tuples.
    sort: bool
        If False, quartets are returned with Nodes spanning edges as
        (below, below, above, above) in idx traversal order given the
        topology and rooting. If sort=True, partitions are instead
        always sorted alphanumerically within and between partitions.
    collapse: bool
        If True then quartets are returned as a single tuple, e.g.,
        (0, 1, 2, 3), else they are returned as a tuple of tuples,
        e.g., ((0, 1), (2, 3)). In either case, the induced split is
        implied to occur in the middle, e.g., 0,1 vs 2,3. Collapse arg
        cannot be combined with type=set.
    quadripartition: bool
        If True then quartets are only returned that are induced by
        quadripartitite splits in a the tree. This is a subset of the
        quartets induced by bipartitions, since the tip Nodes must come
        from four different clades from each edge/split.

    Example
    -------
    >>> tree = toytree.tree("(a,b,((c,d)CD,(e,f)EF)X)AB;")

    >>> # get quartets for each (child, parent) edge in idx order
    >>> list(tree.iter_quartets())
    >>> # [({'c', 'd'}, {'e', 'f'}),
    >>> #  ({'c', 'd'}, {'a', 'f'}),
    >>> #  ({'c', 'd'}, {'b', 'f'}),
    >>> #  ({'c', 'd'}, {'a', 'e'}),
    >>> #  ({'c', 'd'}, {'b', 'e'}),
    >>> #  ({'c', 'd'}, {'a', 'b'}),
    >>> #  ({'e', 'f'}, {'a', 'd'}),
    >>> #  ({'e', 'f'}, {'c', 'd'}),
    >>> #  ({'e', 'f'}, {'b', 'd'}),
    >>> #  ({'e', 'f'}, {'a', 'c'}),
    >>> #  ({'e', 'f'}, {'a', 'b'}),
    >>> #  ({'e', 'f'}, {'b', 'c'}),
    >>> #  ({'d', 'e'}, {'a', 'b'}),
    >>> #  ({'d', 'f'}, {'a', 'b'}),
    >>> #  ({'c', 'd'}, {'a', 'b'}),
    >>> #  ({'e', 'f'}, {'a', 'b'}),
    >>> #  ({'c', 'e'}, {'a', 'b'}),
    >>> #  ({'c', 'f'}, {'a', 'b'})]

    >>> # get same quartets consistently ordered and in simpler format
    >>> sorted(tree.iter_quartets(type=tuple, sort=True, collapse=True))
    >>> # [('a', 'b', 'c', 'd'),
    >>> #  ('a', 'b', 'c', 'd'),
    >>> #  ('a', 'b', 'c', 'e'),
    >>> #  ('a', 'b', 'c', 'f'),
    >>> #  ('a', 'b', 'd', 'e'),
    >>> #  ('a', 'b', 'd', 'f'),
    >>> #  ('a', 'b', 'e', 'f'),
    >>> #  ('a', 'b', 'e', 'f'),
    >>> #  ('a', 'c', 'e', 'f'),
    >>> #  ('a', 'd', 'e', 'f'),
    >>> #  ('a', 'e', 'c', 'd'),
    >>> #  ('a', 'f', 'c', 'd'),
    >>> #  ('b', 'c', 'e', 'f'),
    >>> #  ('b', 'd', 'e', 'f'),
    >>> #  ('b', 'e', 'c', 'd'),
    >>> #  ('b', 'f', 'c', 'd'),
    >>> #  ('c', 'd', 'e', 'f'),
    >>> #  ('c', 'd', 'e', 'f')]
    """
    # disallowed combinations
    if (type == set) and (collapse is True):  # noqa: E721
        collapse = False
        logger.warning(
            "collapse argument cannot be used with type=set, using collapse=False")

    # quartet args
    kwargs = dict(tree=tree, feature=None, quadripartitions=quadripartitions)
    for i, j, x, y in _iter_quartet_sets(**kwargs):

        # sort within by min name
        if type != set:  # noqa: E721
            if i.name > j.name:
                j, i = i, j
            if x.name > y.name:
                x, y = y, x

        # sort between by min name
        if sort:
            if i.name > x.name:
                # print(f"SORTING - {i.name},{x.name} to {x.name},{i.name}")
                i, j, x, y = x, y, i, j

        # convert to feature
        if feature:
            i, j, x, y = (getattr(z, feature) for z in (i, j, x, y))

        # return as collapsed or not.
        if collapse:
            yield type((i, j, x, y))
        else:
            yield type((i, j)), type((x, y))


if __name__ == "__main__":

    import toytree

    tree = toytree.rtree.unittree(6, seed=123, random_names=True)

    print("type=set, collapse=False, sort=False")
    for qrt in iter_quartets(tree, type=set, collapse=False, sort=False):
        print(qrt)
    print("")

    print("type=set, collapse=True, sort=False")
    for qrt in iter_quartets(tree, type=set, collapse=True, sort=False):
        print(qrt)
    print("")

    print("type=tuple, collapse=True, sort=False")
    for qrt in iter_quartets(tree, type=tuple, collapse=True, sort=False):
        print(qrt)
    print("")

    print("\n=======================\n")

    print("type=set, collapse=False, sort=False")
    for qrt in iter_quartets(tree, type=set, collapse=False, quadripartitions=True):
        print(qrt)
    print("")

    print("type=tuple, collapse=False, sort=False")
    for qrt in iter_quartets(tree, type=tuple, collapse=False, quadripartitions=True):
        print(qrt)
    print("")

    print("type=tuple, collapse=True, sort=False")
    for qrt in iter_quartets(tree, type=tuple, collapse=True, quadripartitions=True):
        print(qrt)
    print("")

    print("type=tuple, collapse=False, sort=True")
    for qrt in iter_quartets(tree, type=tuple, collapse=False, sort=True, quadripartitions=True):
        print(qrt)
    print("")

    print("type=tuple, collapse=True, sort=True")
    for qrt in iter_quartets(tree, type=tuple, collapse=True, sort=True, quadripartitions=True):
        print(qrt)
    print("")



#!/usr/bin/env python

"""...

"""

from typing import Tuple, Sequence, Mapping, Any, Union, TypeVar
import numpy as np

from toytree.core import ToyTree, Cartesian, Mark
from toytree.color import ToyColor
from toytree.core.apis import add_subpackage_method, AnnotationAPI
from toytree.annotate.src.checks import get_last_toytree_mark, assert_tree_matches_mark
from toytree.drawing.src.mark_annotation import AnnotationMarker
from toytree.style.src.validate_utils import substyle_dict_to_css_dict
from toytree.style.src.validate_data import (
    validate_colors,
    validate_numeric,
    validate_markers,
    validate_mask,
    validate_labels,
)
from toytree.style.src.validate_nodes import (
    validate_node_style,
)
from toytree.style.src.validate_node_labels import (
    validate_node_labels_style,
)
from toytree.layout.src.get_edge_midpoints import get_edge_midpoints

Color = TypeVar("Color", str, tuple, np.ndarray)

__all__ = [
    "add_edge_markers",
    "add_edge_labels",
    # "add_edge_pie_charts",  # see add_pie_markers.py
]


@add_subpackage_method(AnnotationAPI)
def add_edge_markers(
    tree: ToyTree,
    axes: Cartesian,
    marker: Union[str, Sequence[str]] = "o",
    size: Union[int, Sequence[int]] = 8,
    color: Union[Color, Sequence[Color]] = None,
    opacity: Union[float, Sequence[float]] = 1.0,
    mask: Union[np.ndarray, Tuple[int, int, int], None] = None,
    xshift: float = 0.0,
    yshift: float = 0.0,
    style: Mapping[str, Any] = None,
) -> Mark:
    """Return a toyplot Mark of edge markers added to a tree plot.

    This adds edge markers to the last tree drawn on the Cartesian
    axes. The shape, size, color, and style of markers can be modified.

    Parameters
    ----------
    axes: Cartesian
        A toyplot Cartesian axes object containing a tree drawing.
    marker: str or toyplot.marker.Marker or Sequence
        Marker shape, e.g., "o", "s", "^", "r2x1". See toyplot Markers.
    size: int or Sequence[int]
        Size of markers as single int or Sequence of ints, in px units.
    color: str, tuple, or array, or Sequence
        Color of markers as single color or Sequence of colors.
    opacity: float or Sequence[float]
        Opacity of markers (fill & stroke) as a single float or Sequence
        of floats. Note that fill and stroke opacity can be set
        separately using the style dict, but only as single values.
    mask: np.array or None
        A boolean array of len nnodes or nnodes - 1 where True masks
        an edge from being shown and False shows the edge. None shows
        all edges. A tuple of 3 booleans can be entered as a shortcut
        to build an array of (mask_tips, mask_internal, mask_root).
    style: dict
        Marker style dict. See `tree.style.node_style` for options.
    xshift: int
        Shift marker horizontally by px units (+=right, -=left).
    yshift: int
        Shift marker vertically by px units (+=down, -=up).

    Example
    -------
    >>> tree = toytree.rtree.unittree(6, seed=123)
    >>> canvas, axes, m0 = tree.draw()
    >>> m1 = tree.annotate.add_edge_markers(
    >>>     axes,
    >>>     marker='s',
    >>>     size=9,
    >>>     color='red',
    >>>     style={'stroke': 'white', 'stroke-width': 2.5}
    >>> )
    """
    # get mark for coordinates on plotted tree.
    mark = get_last_toytree_mark(axes)
    assert_tree_matches_mark(tree, mark)

    # only show read edges (not the root edge in unrooted trees)
    nedges = tree.nnodes - 2 if tree.is_rooted() else tree.nnodes - 1

    # get coordinates of all real edges
    # coords = _get_edge_midpoints(tree, mark.ntable, mark.layout, mark.edge_type)[:nedges]
    coords = get_edge_midpoints(mark.etable, mark.ntable, mark.layout, mark.edge_type)[:nedges]

    # mask some edges
    mask = validate_mask(tree, style={"node_mask": mask})[:nedges]

    # set styles on top of defaults. Must run before node_colors.
    style = {} if style is None else style
    style = validate_node_style(tree, style=style, **style)
    style = substyle_dict_to_css_dict(style.__dict__)

    # update node colors setting; sets to None if only one color.
    colors, fill_color = validate_colors(
        tree, key="color", size=tree.nnodes, style={"color": color})

    # if fill_color then set to node_style.fill since node_colors = None
    if colors is None:
        if fill_color:
            style["fill"] = ToyColor(fill_color)  # overrides node_style.fill
        else:
            pass  # node_style.fill overrides
    else:
        colors = colors[:nedges][mask]
        style.pop("fill")

    # ...
    markers = validate_markers(tree, key="edge_markers", size=tree.nnodes, style={"edge_markers": marker})[:nedges][mask]
    sizes = validate_numeric(tree, key="size", size=tree.nnodes, style={"size": size})[:nedges][mask]
    opacity = validate_numeric(tree, key="opacity", size=tree.nnodes, style={"opacity": opacity})[:nedges][mask]

    # logger.warning(coords)
    # logger.warning(mask)
    coords = coords[mask, :]

    # plot edge markers as scatterplot markers
    mark = AnnotationMarker(
        ntable=coords,
        xshift=xshift,
        yshift=yshift,
        sizes=sizes,
        colors=colors,
        opacity=opacity,
        shapes=markers,
        style=style,
    )
    axes.add_mark(mark)
    return mark


@add_subpackage_method(AnnotationAPI)
def add_edge_labels(
    tree: ToyTree,
    axes: Cartesian,
    labels: Union[str, Sequence[str]] = "idx",
    color: Union[Color, Sequence[Color]] = None,
    opacity: Union[float, Sequence[float]] = 1.0,
    font_size: Union[int, None] = 12,
    angle: Union[int, Sequence[int]] = 0,
    mask: Union[np.ndarray, Tuple[int, int, int], None] = None,
    xshift: int = 0,
    yshift: int = 0,
    style: Mapping[str, Any] = None,
) -> Mark:
    """Return a toyplot Mark of edge labels added to a tree drawing.

    This adds edge labels to the last tree drawn on the Cartesian axes.

    Parameters
    ----------
    axes: Cartesian
        A toyplot Cartesian axes object containing a tree drawing.
    labels: str or Sequence[str]
        A sequence of labels as strings in Node idx traversal order.
        The length must be equal to nnodes or nnodes - 1. Use '' to
        not add a label to some edges. You can also enter a single str
        as the name of a feature on the tree to extract data as labels,
        e.g., "idx" will show int idx labels for edges.
    color: str, tuple, array or Sequence
        A single color or Sequence of colors for edge labels.
    opacity: float or Sequence[float]
        A single opacity or Sequence of opacities for edge labels.
    font_size: float
        Font size in px. Overrides 'font-size' setting in style dict.
    angle: int or Sequence[int]
        A single angle applied to all labels, or Sequence of angles.
    mask: np.array or None
        A boolean array of len nnodes or nnodes - 1 where True masks
        an edge from being shown and False shows the edge. None shows
        all edges. A tuple of 3 booleans can be entered as a shortcut
        to build an array of (mask_tips, mask_internal, mask_root).
    xshift: int
        Shift label horizontally by px units (+=right, -=left).
    yshift: int
        Shift label vertically by px units (+=down, -=up).
    style: dict
        Style dict. See `tree.style.node_labels_style` for options.

    Example
    -------
    >>> tree = toytree.rtree.unittree(6, seed=123)
    >>> canvas, axes, m0 = tree.draw()
    >>> m1 = tree.annotate.add_edge_labels(
    >>>     axes,
    >>>     labels=tree.get_node_data("idx"),
    >>>     color='blue',
    >>>     style={'font-size': 16, 'baseline-shift': 8}
    >>> )
    """
    # get mark for coordinates on plotted tree.
    mark = get_last_toytree_mark(axes)
    assert_tree_matches_mark(tree, mark)

    # get coordinates of all real edges
    nedges = tree.nnodes - 2 if tree.is_rooted() else tree.nnodes - 1
    # coords = _get_edge_midpoints(tree, mark.ntable, mark.layout, mark.edge_type)[:nedges]
    coords = get_edge_midpoints(mark.etable, mark.ntable, mark.layout, mark.edge_type)[:nedges]

    # mask some edges
    mask = validate_mask(tree, style={"node_mask": mask})[:nedges]
    labels = validate_labels(
        tree, key="labels", size=tree.nnodes, style={"labels": labels})[:nedges][mask]

    # set styles on top of defaults
    style = {} if style is None else style
    style = validate_node_labels_style(tree, style=None, **style)
    style = substyle_dict_to_css_dict(style.__dict__)

    # override font size
    if font_size:
        style["font-size"] = font_size

    # check colors
    label_colors, fill_color = validate_colors(
        tree, key="color", size=tree.nnodes, style={"color": color})
    if label_colors is None:
        if fill_color:
            style["fill"] = ToyColor(fill_color)  # overrides ..._style.fill
        else:
            pass  # node_style.fill overrides
    else:
        label_colors = label_colors[:nedges][mask]
        style.pop("fill")

    # mask some nodes
    opacity = validate_numeric(
        tree, key="opacity", size=tree.nnodes, style={"opacity": opacity})[:nedges][mask]
    angle = validate_numeric(
        tree, key="angle", size=tree.nnodes, style={"angle": angle})[:nedges][mask]

    # expand xshift,yshift args as anchor_shift,baseline_shift
    style['-toyplot-anchor-shift'] += xshift
    style['baseline-shift'] -= yshift

    # add text at Node positions + half length of dists.
    coords = coords[mask, :]
    mark = axes.text(
        coords[:, 0],
        coords[:, 1],
        labels,
        color=label_colors,
        opacity=opacity,
        angle=angle,
        style=style,
        annotation=True,
    )
    return mark


if __name__ == "__main__":

    import toytree

    # base tree drawing
    tree = toytree.rtree.unittree(6)  # .unroot()
    tree[0].name = "amdodfl"
    tree[1]._name = "HI"
    c, a, m = tree.draw(layout='d')  # r')

    # annotate with edge labels
    add_edge_markers(tree, axes=a, size=10, marker='r1x2', color="height")
    add_edge_labels(tree, axes=a, labels="idx", font_size=15)

    data = np.array([[0.5, 0.3, 0.2]] * tree.nnodes)
    toytree.utils.show(c, tmpdir="~")



#!/usr/bin/env python

"""Annotation methods for adding tip markers to tree drawings.

Examples
--------
...
"""

from typing import Tuple, Sequence, Mapping, Any, Union, TypeVar
import numpy as np
from toyplot.mark import Mark
from toyplot.coordinates import Cartesian

from toytree import ToyTree
from toytree.color import ToyColor
from toytree.core.apis import add_subpackage_method, AnnotationAPI
from toytree.style.src.validate_utils import substyle_dict_to_css_dict
from toytree.annotate.src.checks import get_last_toytree_mark, assert_tree_matches_mark

# WORK IN PROGRESS...
from toytree.drawing.src.mark_annotation import AnnotationMarker
from toytree.style.src.validate_data import (
    validate_colors,
    validate_markers,
    validate_numeric,
    validate_labels,
)
from toytree.style.src.validate_node_labels import validate_node_labels_style
from toytree.style.src.validate_nodes import (
    validate_node_style,
)

Color = TypeVar("Color", str, tuple, np.ndarray)

__all__ = [
    # "add_tip_labels",
    "add_tip_markers",
    # "add_tip_pie_charts",
    # "add_node_bars",
    # "add_node_histograms",
    # "add_node_densigrams",
]


@add_subpackage_method(AnnotationAPI)
def add_tip_markers(
    tree: ToyTree,
    axes: Cartesian,
    align: bool = None,  # default=None and use T/F to override ts?
    marker: Union[str, Sequence[str]] = "o",
    size: Union[int, Sequence[int]] = 8,
    color: Union[Color, Sequence[Color]] = None,
    opacity: Union[float, Sequence[float]] = 1.0,
    mask: Union[np.ndarray, Tuple[int, int, int], bool, None] = None,
    style: Mapping[str, Any] = None,
    xshift: int = 0,
    yshift: int = 0,
    # label: Optional[str] = None,
    # label_angle: int = 45,
) -> Mark:
    """Return toyplot Mark of markers aligned with tips of tree plot.

    This adds node markers to the last tree drawn on the Cartesian
    axes using the coordinates of plotted Nodes. The shape, size,
    color, and style of markers can be modified.

    Parameters
    ----------
    axes: Cartesian
        A toyplot Cartesian axes object containing a tree drawing.
    align: bool
        If True markers align at the farthest tip position; if False
        markers align at each tip's position; if None markers align
        according to the tip_labels_align arg used during tree drawing.
    marker: str or toyplot.marker.Marker or Sequence
        Marker shape, e.g., "o", "s", "^", "r2x1". See toyplot Markers.
    size: int or Sequence[int]
        Size of markers as single int or Sequence of ints, in px units.
    color: str, tuple, or array, or Sequence
        Color of markers as single color or Sequence of colors.
    opacity: float or Sequence[float]
        Opacity of markers (fill & stroke) as a float or Sequence
        of floats. Note: fill and stroke opacities can be set
        separately using the style dict, but only as single values.
    style: dict
        Marker style dict. See `tree.style.node_style` for options.
    xshift: int
        Shift position of markers by x canvas (px) units on x axis.
    yshift: int
        Shift position of markers by x canvas (px) units on y axis.

    Example
    -------
    >>> tree = toytree.rtree.unittree(6, seed=123)
    >>> canvas, axes, m0 = tree.draw()
    >>> # add markers to all Nodes
    >>> m1 = tree.annotate.add_node_markers(
    >>>     axes,
    >>>     marker='s',
    >>>     size=9,
    >>>     color='red',
    >>>     style={'stroke': 'white', 'stroke-width': 2.5}
    >>> )
    >>>
    >>> # add markers to only a few Nodes
    >>> m2 = tree.annotate.add_node_markers(
    >>>     axes, marker=">", size=20, mask=tree.get_node_mask(9)
    >>> )
    """
    # get mark for coordinates on plotted tree.
    mark = get_last_toytree_mark(axes)
    assert_tree_matches_mark(tree, mark)

    # get coordinates of the tip Nodes
    coords = mark.ttable.copy()

    # align tips: if mark.tip_labels_align == True this is already done
    align = align if align is not None else mark.tip_labels_align
    if align:
        if mark.layout == "r":
            coords[:, 0] = coords[:, 0].max()
        elif mark.layout == "l":
            coords[:, 0] = coords[:, 0].min()
        elif mark.layout == "u":
            coords[:, 1] = coords[:, 1].max()
        elif mark.layout == "d":
            coords[:, 1] = coords[:, 1].min()
        else:
            raise NotImplementedError("TODO")

    # update node_style setting
    style = {} if style is None else style
    style = validate_node_style(tree, style=None, **style)
    style = substyle_dict_to_css_dict(style.__dict__)

    # update node colors setting; sets to None if only one color.
    colors, fill_color = validate_colors(
        tree, key="colors", size=tree.ntips, style={"colors": color},
    )

    # if fill_color then set to node_style.fill since node_colors = None
    if colors is None:
        if fill_color:
            style["fill"] = ToyColor(fill_color)  # overrides node_style.fill
        else:
            pass  # node_style.fill overrides
    else:
        colors = colors
        style.pop("fill")

    # validate others and trim to mask
    markers = validate_markers(
        tree, key="node_markers", size=tree.ntips, style={"node_markers": marker})
    sizes = validate_numeric(
        tree, key="size", size=tree.ntips, style={"size": size})
    opacity = validate_numeric(
        tree, key="opacity", size=tree.ntips, style={"opacity": opacity})

    # create custom Mark that allows for [xy]shift
    mark = AnnotationMarker(
        ntable=coords,
        xshift=xshift,
        yshift=yshift,
        sizes=sizes,
        colors=colors,
        opacity=opacity,
        shapes=markers,
        style=style,
    )
    axes.add_mark(mark)
    return mark

    # add label for the data column
    # if label:
    #     label = axes.text(
    #         coords[0, 0] + xoffset,
    #         tree.ntips,
    #         label,
    #         angle=label_angle,
    #         style={'text-anchor': 'start', 'fill': "#252525"},
    #     )


# TODO: auto-angles for different layouts...
# @add_subpackage_method(AnnotationAPI)
# def add_tip_labels(
#     tree: ToyTree,
#     axes: Cartesian,
#     labels: Union[str, Sequence[str]] = "idx",
#     color: Union[Color, Sequence[Color]] = None,
#     opacity: Union[float, Sequence[float]] = 1.0,
#     font_size: Union[int, None] = 12,
#     angle: Union[int, Sequence[int]] = 0,
#     offset: int = 15,
#     style: Mapping[str, Any] = None,
# ) -> Mark:
#     """Return a toyplot Mark of node labels added to a tree drawing.

#     This adds node labels to the last tree drawn on the Cartesian axes.

#     Parameters
#     ----------
#     axes: Cartesian
#         A toyplot Cartesian axes object containing a tree drawing.
#     labels: str or Sequence[str] or None
#         A sequence of labels as strings in Node idx traversal order.
#         The length must be equal to nnodes or nnodes - 1. Use '' to
#         not add a label to some edges. You can also enter a single str
#         as the name of a feature on the tree to extract data as labels,
#         e.g., "idx" will show int idx labels for nodes.
#     color: str, tuple, array or Sequence
#         A single color or Sequence of colors for node labels.
#     opacity: float or Sequence[float]
#         A single opacity or Sequence of opacities for node labels.
#     font_size: float
#         Font size in px. Overrides 'font-size' setting in style dict.
#     angle: int or Sequence[int]
#         A single angle applied to all labels, or Sequence of angles.
#     mask: np.array or None
#         A boolean array of len nnodes or nnodes - 1 where True masks
#         an node from being shown and False shows the node. None or False
#         shows all nodes. A tuple of 3 booleans can be entered as a
#         shortcut to (show_tips, show_internal, show_root).
#     style: dict
#         Style dict. See `tree.style.node_labels_style` for options.

#     Example
#     -------
#     >>> tree = toytree.rtree.unittree(6, seed=123)
#     >>> canvas, axes, m0 = tree.draw()
#     >>> m1 = tree.annotate.add_node_labels(
#     >>>     axes,
#     >>>     labels=tree.get_node_data("idx"),
#     >>>     color='blue',
#     >>>     style={'font-size': 16, 'baseline-shift': 8}
#     >>> )
#     """
#     # get mark for coordinates on plotted tree.
#     mark = get_last_toytree_mark_from_cartesian(axes)
#     assert_tree_matches_mark(tree, mark)
#     coords = mark.ttable

#     # check length and type of labels
#     labels = validate_tip_labels(tree, style=None, tip_labels=labels)

#     # set styles on top of defaults
#     style = {"-toyplot-anchor-shift": 0} if style is None else style
#     style = validate_tip_labels_style(tree, style=None, **style)
#     style = substyle_dict_to_css_dict(style.__dict__)

#     # override font size
#     if font_size:
#         style["font-size"] = font_size

#     # update node colors setting; sets to None if only one color.
#     node_colors, fill_color = validate_tip_labels_colors(
#         tree, style=None, tip_labels_colors=color)

#     # if fill_color then set to node_style.fill since node_colors = None
#     if node_colors is None:
#         if fill_color:
#             style["fill"] = ToyColor(fill_color)  # overrides node_style.fill
#         else:
#             pass  # node_style.fill overrides
#     else:
#         node_colors = node_colors
#         style.pop("fill")

#     # opacity = validate_node_numeric(tree, style=None, key="opacity", opacity=opacity)
#     # angle = validate_node_numeric(tree, style=None, key="angle", angle=angle)
#     style['fill-opacity'] = 1.0

#     # add text at Node positions + half length of dists.
#     mark = axes.text(
#         coords[:, 0],
#         coords[:, 1],
#         labels,
#         color=node_colors,
#         opacity=opacity,
#         angle=angle,
#         style=style,
#         annotation=True,
#     )
#     return mark


# @add_subpackage_method(AnnotationAPI)
# def add_node_pie_charts(
#     tree: ToyTree,
#     axes: Cartesian,
#     data: np.ndarray,
#     size: Union[int, Sequence[int]] = 10,
#     colors: Union[Sequence[Color], Color] = None,
#     ostroke: Color = "#262626",
#     ostroke_width: float = 1.5,
#     istroke: Color = "#262626",
#     istroke_width: float = 0.,
#     rotate: int = -45,
#     mask: Union[bool, np.ndarray, tuple] = False,
#     xshift: int = 0,
#     yshift: int = 0,
# ) -> Mark:
#     """Return a toyplot Mark of node markers added to a tree plot.

#     This adds node markers to the last tree drawn on the Cartesian
#     axes. The shape, size, color, and style of markers can be modified.

#     Parameters
#     ----------
#     axes: Cartesian
#         A toyplot Cartesian axes object containing a tree drawing.
#     data: numpy.ndarray
#         Array of shape(ncategories, nnodes) with rows summing to 1. 
#     size: int or Sequence[int]
#         Size of markers as single int or Sequence of ints, in px units.
#     colors: None, str, tuple, or array, or Sequence
#         Color for each category/trait or the name of a colormap.
#     ostroke: Color
#         Color of the stroke on the outside of the Mark.
#     ostroke_width: float
#         Width of the stroke on the outside of the Mark
#     istroke: Color
#         Color of the stroke on the inside of the Mark between wedges.
#     istroke_width: float
#         Width of the stroke on the inside of the Mark between wedges.
#     rotate: int
#         Rotate the starting point of the wedges.
#     mask: bool, np.ndarray, or tuple
#         Node mask to hide/show some or all Nodes.
#     xshift: int
#         Shift marker horizontally by px units (+=right, -=left).
#     yshift: int
#         Shift marker vertically by px units (+=down, -=up).

#     Example
#     -------
#     >>> tree = toytree.rtree.unittree(6, seed=123)
#     >>> canvas, axes, m0 = tree.draw()

#     >>> # generate random pie-like (proportion) data array
#     >>> import numpy as np
#     >>> ncategories = 3
#     >>> arr = np.random.random(size=(tree.nnodes, ncategories))
#     >>> arr = (arr.T / arr.sum(axis=1)).T

#     >>> # add pie charts to all internal Nodes
#     >>> tree.annotate.add_node_pie_charts(
#     >>>     axes=axes, data=arr, size=20, mask=(0, 1, 1),
#     >>>     istroke_width=0.75, istroke="black", rotate=-45,
#     >>> )
#     """
#     # get mark for coordinates on plotted tree.
#     mark = get_last_toytree_mark_from_cartesian(axes)
#     assert_tree_matches_mark(tree, mark)

#     # get mask
#     mask = validate_node_mask(tree, style=None, node_mask=mask)

#     # check and cleanup data input.
#     # TODO: option to collapse categories below a minimum percentage
#     data = validate_pie_data(tree, data)

#     # expand colormap to an array of colors
#     if colors is None:
#         colors = "Set2"
#     if isinstance(colors, (tuple, list, np.ndarray)):
#         pass
#     else:
#         colors = get_color_mapped_values(range(data.shape[1]), colors)

#     # ensure conversion of colors to array type and size=ncategories
#     colors = check_arr(
#         values=colors,
#         label="colors (pie chart category colors)",
#         size=data.shape[1],
#         ctype=np.void,
#     )
#     sizes = validate_node_sizes(tree, style=None, node_sizes=size)[mask]

#     # mask some Nodes
#     data = data[mask, :]
#     coords = mark.ntable[mask, :]
#     if xshift or yshift:
#         # Note: if later annotations change the projection this will be off
#         axes._finalize()
#         if xshift:
#             origin, xshifted = axes._x_projection.inverse([0, xshift])
#             x_shift_projected = xshifted - origin
#             coords[:, 0] += x_shift_projected
#         if yshift:
#             origin, yshifted = axes._y_projection.inverse([0, yshift])
#             y_shift_projected = yshifted - origin
#             coords[:, 1] += y_shift_projected
#         axes._finalized = None

#     # plot edge markers as scatterplot markers
#     mark = PieChartMark(
#         coordinates=coords,
#         data=data,
#         sizes=sizes,
#         colors=colors,
#         ostroke=ostroke,
#         ostroke_width=ostroke_width,
#         istroke=istroke,
#         istroke_width=istroke_width,
#         rotate=rotate,
#     )
#     axes.add_mark(mark)
#     return mark


# def add_node_bars(
#     tree: ToyTree,
#     axes: Cartesian,
#     # bars: Mapping[Query, Tuple[float, float]],
#     color: Union[Color, Sequence[Color]],
#     opacity: Union[float, Sequence[float]],
# ):
#     """

#     Example
#     -------
#     >>> tree = toytree.rtree.unittree(10, treeheight=1e5)
#     >>> ages_ci = {
#     >>>     nidx: (node.dist - 1000, node.dist + 1000)
#     >>>     for nidx, node in enumerate(tree)
#     >>> }
#     >>> canvas, axes, mark0 = tree.draw()
#     >>> tree.annotate.node_height_confidence_intervals(axes, ages_ci)
#     >>> 
#     >>> mark = toytree.annotate.node_height_confidence_intervals(
#     >>>     tree=tree, axes=axes, mapping=ages_ci)
#     """
#     pass


# def add_clade_box(
#     tree: ToyTree,
#     axes: Cartesian,
# ):
#     pass


if __name__ == "__main__":

    import toytree

    # base tree drawing
    tree = toytree.rtree.unittree(6)
    c, a, m = tree.draw(layout='d')

    # annotate with edge labels
    # add_edge_labels(tree, axes=a, labels=tree.get_node_data("idx"))
    # add_node_markers(
    #     tree, axes=a, size=10, marker='s', color=("idx", "BlueRed"))
    # add_node_labels(tree, axes=a, labels='idx')
    data = np.array([[0.5, 0.3, 0.2]] * tree.nnodes)
    m = tree.annotate.add_tip_markers(a, color=("name"), xshift=0, yshift=50)
    print(m.extents(['x', 'y']))
    toytree.utils.show(c)



#!/usr/bin/env python

"""Random or fixed tree generation subpackage

The :mod:`toytree.core.random.rtree` is accessible from the top-level
of the toytree package as :mod:`toytree.rtree`. This module includes
functions for generating random trees, like `rtree` and `bdtree`, or
fixed trees of topology shapes that are often useful for research,
such as `baltree` or `imbtree`. 

Examples
--------
>>> tree1 = toytree.rtree.rtree(10)
>>> tree2 = toytree.rtree.unittree(ntips=10, seed=123)
>>> tree3 = toytree.rtree.baltree(ntips=10, treeheight=1e6)
"""

from toytree.rtree._src.rtree import (
	rtree, unittree, imbtree, baltree, bdtree, coaltree)



#!/usr/bin/env python

"""Tests using validate_style to test draw/style args.

"""

import unittest
import numpy as np
# import toyplot
import toytree
# from toytree.color import ToyColor
from toytree.style import TreeStyle, validate_style
from toytree.utils import ToytreeError
from toytree.style.src.validate_data import (
    validate_mask,
    validate_numeric,
    validate_markers,
    validate_colors,
    validate_hover,
    validate_labels,
    validate_admixture_edges,
)



class TestValidateStyle(unittest.TestCase):

    def setUp(self):
        self.tree = toytree.rtree.unittree(10, seed=123)
        self.style = TreeStyle()
        self.cycler = toytree.color.color_cycler()

    def test_style_copy(self):
        """Copying a style should deepcopy (copies of substyles and dicts)."""
        _ = self.tree.style.copy()


    def test_style_raise_exception_on_validate_bad_color(self):
        """show style in notebook without fail on complex settings."""
        self.style.node_colors = "..."
        with self.assertRaises(ToytreeError):
            validate_style(self.tree, self.style)


class TestValidateMask(unittest.TestCase):

    def setUp(self):
        self.tree = toytree.rtree.unittree(5, seed=123)

    def test_style_validate_mask(self):
        tests = [
            True,
            False,
            np.zeros(self.tree.nnodes, dtype=np.bool_),        
            np.ones(self.tree.nnodes, dtype=np.bool_),
            self.tree.get_node_mask(),
            self.tree.get_node_mask(2, 3, 4),
            self.tree.get_node_mask(2, 3, 4, show_tips=True),
            self.tree.get_node_mask(2, 3, 4, show_internal=True),
            self.tree.get_node_mask(2, 3, 4, show_root=True),
            (1, 1, 0),
            (0, 0, 0),
        ]
        for arg in tests:
            validate_mask(
                tree=self.tree, tree_style=self.tree.style,
                style={"node_mask": arg})

    def test_style_raise_on_bad_mask(self):
        """..."""


class TestValidateNumeric(unittest.TestCase):

    def setUp(self):
        self.tree = toytree.rtree.unittree(5, seed=123)

    def test_style_validate_numeric_int(self):
        tree = self.tree.copy()
        tree.style.node_sizes = 10
        vals = validate_numeric(
            tree=tree,
            key="node_sizes",
            size=tree.nnodes,
            tree_style=tree.style,
            style={},
        )
        self.assertTrue(isinstance(vals, np.ndarray))
        self.assertEqual(vals[0], 10)
        self.assertTrue(isinstance(vals[0], np.integer))
        self.assertTrue(all(i == 10 for i in vals))

    def test_style_validate_numeric_int_array(self):
        tree = self.tree.copy()
        tree.style.node_sizes = np.arange(tree.nnodes)
        vals = validate_numeric(
            tree=tree,
            key="node_sizes",
            size=tree.nnodes,
            tree_style=tree.style,
            style={},
        )
        self.assertTrue(isinstance(vals, np.ndarray))
        self.assertEqual(vals[1], 1)
        self.assertTrue(isinstance(vals[1], np.integer))

    def test_style_validate_numeric_float_with_nan(self):
        tree = self.tree.copy()
        tree.style.node_sizes = np.ones(tree.nnodes)
        tree.style.node_sizes[3] = np.nan
        vals = validate_numeric(
            tree=tree,
            key="node_sizes",
            size=tree.nnodes,
            tree_style=tree.style,
            style={},
        )
        self.assertTrue(isinstance(vals, np.ndarray))
        self.assertEqual(vals[1], 1.0)
        self.assertTrue(isinstance(vals[1], np.float64))
        self.assertTrue(np.isnan(vals[3]))

    def test_style_validate_numeric_none(self):
        """None should return array of 0 of correct length"""
        tree = self.tree.copy()
        tree.style.node_sizes = None
        vals = validate_numeric(
            tree=tree,
            key="node_sizes",
            size=tree.nnodes,
            tree_style=tree.style,
            style={},
        )
        self.assertTrue(isinstance(vals, np.ndarray))
        self.assertEqual(vals[0], 0.0)
        self.assertTrue(isinstance(vals[1], np.integer))

    # def test_style_validate_numeric_tuple_format(self):
    #     tree = self.tree.copy()
    #     for test in [
    #         "dist",                 # (dist, 5, 20)
    #         ("dist",),              # (dist, 5, 20)
    #         # ("dist", 2),          # (dist, 2, 20)
    #         # ("dist", 2, 5),       # (dist, 2, 5)
    #         # ("dist", 2, 5, 0),    # (dist, 2, 5, 0)
    #         ]:
    #         tree.style.node_sizes = test
    #         vals = validate_numeric(
    #             tree=tree,
    #             key="node_sizes",
    #             size=tree.nnodes,
    #             tree_style=tree.style,
    #             style={},
    #         )
    #         self.assertTrue(isinstance(vals, np.ndarray))
    #         self.assertEqual(vals[0], 1 / 3.)
    #         self.assertTrue(isinstance(vals[1], np.float64))



    def test_style_validate_numeric_nnodes(self):
        self.tree.style.node_sizes = 5.
        tests = [
            None,  # gets it from .style instead of from arg
            np.full(self.tree.nnodes, 5.),
            np.arange(self.tree.nnodes, dtype=np.float64),
            5.,
            [0, 1, 2, 3, 4, 5, 6, 7, 8],
            [0, 1, 2, 3, 4, 5., 6., 7., 8.],
            [0, 1, 2, 3, 4, 5., 6., np.nan, np.nan],
            ('dist',),
            ('dist', 0, 5),
            ('support', 0, 5, 5.),
        ]
        for arg in tests:
            vals = validate_numeric(
                tree=self.tree,
                key="node_sizes",
                tree_style=self.tree.style,
                size=self.tree.nnodes,
                style={"node_sizes": arg, "other": '...'})
            print(vals)
            self.assertTrue(len(vals) == self.tree.nnodes)
            self.assertTrue(len(vals) == self.tree.nnodes)            





    # def test_style_validate_multi_colors(self):
    #     """show style in notebook without fail on complex settings."""
    #     self.style.node_colors = ["red"] + ["blue"] * (self.tree.nnodes - 1)
    #     self.style.tip_labels_colors = [next(self.cycler) for i in range(self.tree.ntips)]
    #     # self.style.node_labels_style.stroke = ToyColor((1, 0, 0.5, 0.5))
    #     # self.style.edge_colors = (1, 0, 0.5, 0.5)
    #     validate_style(self.tree, self.style)



if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

"""Apply styling to toyplot Cartesian axes objects.

Set margin to 60.
"""

import toyplot


def set_axes_ticks_external(axes: toyplot.coordinates.Cartesian) -> toyplot.coordinates.Cartesian:
    """Add custom generic style to plot axes.

    Sets tick marks to extend 5px outside of spine, tick mark labels
    to 10px outside of spine, and text labels to 30px outside of spine.
    To make sure that labels can still fit, the margin is also increased
    by 10px, so that it is 60px on top, bottom, left and right.
    """
    axes.x.ticks.show = True
    axes.x.domain.show = False
    axes.y.domain.show = False
    axes.y.ticks.show = True
    axes.x.ticks.near = 5
    axes.x.ticks.far = 0
    axes.y.ticks.near = 5
    axes.y.ticks.far = 0
    axes.x.ticks.labels.offset = 10
    axes.y.ticks.labels.offset = 10
    axes.y.label.offset = 30
    axes.x.label.offset = 30
    axes.label.offset = 20

    # this is equivalent to increasing margin from default=50 to 60.
    axes._xmin_range += 10
    axes._xmax_range -= 10
    axes._ymin_range += 10
    axes._ymax_range -= 10
    return axes


def set_axes_box_outline(axes: toyplot.coordinates.Cartesian) -> toyplot.coordinates.Cartesian:
    """Adds a box outline to plots.

    This is achieved by creating new axes objects that ignore the data
    domain and extend their spines all the way to the extent of the
    axes padding.

    Note
    ----
    To easily set the stroke-width of ALL spines, ticks, etc you can
    use the canvas.style['stroke-width'] setting. Also, note that if
    the axes.padding is changed AFTER this function is called then the
    axes can be separated from the box outline (i.e., this func uses
    the curretn padding value to align the two).
    """
    # opposite of original y
    newy = axes.share("x", xlabel=axes.x.label.text)
    # newy.y.spine.style['stroke'] = stroke
    newy.y.ticks.show = False
    newy.y.ticks.labels.show = False
    newy._ymin_range -= axes.padding
    newy._ymax_range += axes.padding

    # on top of original y, inherits the ymin,max range from newy
    newy = newy.share("x", xlabel=axes.x.label.text)
    # newy.y.spine.style['stroke'] = stroke
    newy.y.spine.position = "low"
    newy.y.ticks.show = False
    newy.y.ticks.labels.show = False

    # opposite of original x
    newx = axes.share("y", ylabel=axes.y.label.text)
    # newx.x.spine.style['stroke'] = stroke
    newx.x.ticks.show = False
    newx.x.ticks.labels.show = False
    newx._xmin_range -= axes.padding
    newx._xmax_range += axes.padding

    # on top of original x
    newx = newx.share("y", ylabel=axes.y.label.text)
    # newx.x.spine.style['stroke'] = stroke
    newx.x.spine.position = "low"
    newx.x.ticks.show = False
    newx.x.ticks.labels.show = False
    return axes


if __name__ == "__main__":

    import toytree
    import numpy as np

    # ...
    tree = toytree.rtree.bdtree(50, b=5, d=2.5)
    c1, a, m = tree.draw(scale_bar=True, width=300, height=225, tip_labels=False, padding=10)
    a.x.label.text = "hello world"
    a.y.label.text = "hello world"
    set_axes_ticks_external(a)
    set_axes_box_outline(a)
    c1.style['stroke-width'] = 1.5

    # ...
    heights = np.array(sorted([-i.height for i in tree.traverse() if not i.is_leaf()]))
    loglineages = np.log(np.arange(2, tree.ntips + 1))
    c2, a, m = toyplot.plot(
        heights, loglineages,
        width=300,
        height=225,
        stroke_width=3,
        xlabel="time",
        ylabel="log n lineages",
    )
    set_axes_ticks_external(a)
    set_axes_box_outline(a)
    c2.style['stroke'] = 'grey'
    c2.style['stroke-width'] = 1.5
    toytree.utils.show([c1, c2])



#!/usr/bin/env python

"""Enumeration counting module.

This module implements mathematical operations related to counting
numbers of trees and metrics related to tree space.

Authors
-------
Deren Eaton and Carlos Alonso Maya-Lastra

TODO
----
- should all funcs take int|ToyTree as input and be exposed to Tree API?
- check and explain n labeled/unlabeld versus rooted/unrooted trees.

References
----------
- Edwards AWF (1970). Estimation of the branch points of a branching diffusion process. J. R. Stat. Soc. Ser. B 32:155-174
- Harding EF (1971) The probabilities of rooted tree-shapes generated by random bifurcation. Adv Appl Prob 3: 44–77.
Harding provides recursive formula for hte probabilities of labeled and unlabeled topologies. See Brown for full solutions.
- Brown JKM (1994) Probabilities of evolutionary trees. Syst Biol 43: 78–91.
Brown developed probability calculations for all labeled and unlabeled topologies by developing expressions for the number of ways that internal nodes can be arranged relative to one another in time.
- Aldous DJ (2001) Stochastic models and descriptive statistics for phylogenetic trees, from Yule to today. Stat Sci 16: 23–34.
- Steel M, McKenzie A (2001) Properties of phylogenetic trees generated by Yule-type speciation models. Math Biosci 170: 91–112.
- Pamilo and Nei (1988) ...
"""

from math import prod, factorial
# import numpy as np
from scipy.special import comb as scipy_comb
from toytree.core import ToyTree
from toytree.core.apis import TreeEnumAPI, add_subpackage_method


__all__ = [
    "get_num_bifurcating_trees",     # Module
    # "get_num_multifurcating_trees",  # Module
    "get_num_quartets",              # Module and Tree API
    "get_num_subtrees",              # Module and Tree API
    "get_num_labeled_trees",         # Module
    # "get_num_unlabeled_trees",
    # "iter_labeled_trees",    # moved to enum/topologies
    # "iter_unlabeled_trees",  # moved to enum/topologies
]


# -------------------------------------------------------------------
# -------------------------------------------------------------------
# Didactic functions used in lessons. See `get_num_bifurcating_trees`
# as the recommended faster option for counting. However, this slower
# approach allows us to also accommodate the case of multifurcating
# trees. For this see `get_num_trees_with_multifurcations`.
def _get_num_places_a_tip_can_be_added(ntips: int, rooted: bool = True) -> int:
    """Return num places a tip can be added to a bifurcating rooted
    or unrooted tree.

    Rooted = `2 * (ntips + 1) - 3`.
    Unrooted = `2 * (ntips + 1) - 5`
    """
    if not rooted:
        return 2 * (ntips + 1) - 5
    return 2 * (ntips + 1) - 3


def _get_num_bifurcating_trees_by_successive_odds(ntips: int, rooted: bool = True) -> int:
    """Return number of rooted trees

    As discussed in Felsenstein (2004), there are a couple of ways to
    get the possible number of bifurcating rooted trees considering
    the places where a n-tip can be added. Considering one of the
    methods mentioned, we can create a list of successive odds
    (places where a new tip can be added) and use it to get the
    number of possible tips.
    """
    # To save list of successive odd integers
    successive_odds = (
        _get_num_places_a_tip_can_be_added(i, rooted=rooted)
        for i in range(2, ntips)
    )
    return prod(successive_odds)


# TODO: NOT YET TESTED. REVISIT FELSENSTEIN
# def get_num_multifurcating_trees(
#     ntips: int,
#     n_internal_nodes: int,
#     rooted: bool = True,
# ) -> int:
#     """Return the number of trees...

#     Following Felsenstein ...
#     """
#     # reduce one possible tip in unrooted trees
#     ntips = ntips if rooted else ntips - 1

#     # calculate n_nodes to create matrix
#     nnodes = ntips - 1

#     # create an empty numpy array to store the progressive results
#     matrix_ntrees = np.zeros(shape=(nnodes + 1, ntips + 1))

#     # iterate over each n and each m
#     for m in range(1, nnodes + 1):
#         for n in range(2, ntips + 1):

#             # fill diagonal that is equal to the bifurcating trees
#             if m == n - 1:
#                 matrix_ntrees[m, n] = (
#                     _get_num_bifurcating_trees_by_successive_odds(ntips=n, rooted=rooted))

#             # if it is a different m than n-1, fill the multifurcating trees counts
#             else:
#                 # m = 1 is an especial case because the number of trees is always 1
#                 if m == 1:
#                     matrix_ntrees[m, n] = 1

#                 # if not m =1 them apply the following formula
#                 else:
#                     val = (n + m - 2) * matrix_ntrees[m-1, n-1] + m * matrix_ntrees[m, n-1]
#                     matrix_ntrees[m, n] = val

#     # if n_internal_nodes is passed get position in matrix that store n_tips and n_internal_nodes and store it as int
#     if n_internal_nodes:
#         result = int(matrix_ntrees[n_internal_nodes, ntips])
#     else:
#         # get max multifurcating trees
#         result = int(np.sum(matrix_ntrees[:, ntips]))
#     return result

# -------------------------------------------------------------------
# -------------------------------------------------------------------


def get_num_bifurcating_trees(ntips: int, rooted: bool = True) -> int:
    """Return number of bifurcating trees for ntips (leaf nodes) as
    rooted or unrooted trees.

    Equations
    ---------
    Rooted = (2x-3)! / (2**(x-2) (x-3)!)
    Unrooted = (2 (x-1) -3)! / (2**[x-3] (x-3)!)

    Parameters
    ----------
    ntips: int
        The number of tips.
    rooted: bool
        If True the equation for counting rooted trees is used else
        the equation for unrooted trees. There are always more possible
        rooted trees than unrooted trees.
    """
    # rooted case
    if rooted:
        ntrees = int(
            factorial(2 * ntips - 3)
            / (2 ** (ntips - 2) * factorial(ntips - 2))
        )
    # unrooted case
    else:
        ntrees = int(
            factorial(2 * (ntips - 1) - 3)
            / (2 ** (ntips - 3) * factorial(ntips - 3))
        )
    return ntrees


@add_subpackage_method(TreeEnumAPI)
def get_num_quartets(ntips: int | ToyTree, method: int = 0) -> int:
    """Return number of possible quartet trees (ntips=4) that can be
    extracted from a larger tree of size ntips.

    This uses the formula for binomial coefficient.

    Parameters
    ----------
    ntips: int
        The number of tips in the full tree.

    Note
    ----
    This is a special case of the `enum.get_num_subtrees` function
    which can calculate the number of subtrees of size >4.
    """
    # input can be int or a ToyTree from which ntips is extracted
    ntips = ntips if isinstance(ntips, int) else ntips.ntips

    # fastest method tested
    return scipy_comb(ntips, 4, exact=True)


@add_subpackage_method(TreeEnumAPI)
def get_num_subtrees(ntips: int | ToyTree, subtree_size: int) -> int:
    """Return number of possible subtrees of subtree_size tips that
    can be extracted from a larger tree of size ntips.

    This uses the formula for binomial coefficient.

    Parameters
    ----------
    ntips: int
        The number of tips in the full tree.
    subtree_size: int
        The number of tips in the subtrees (subtree size).
    """
    # input can be int or a ToyTree from which ntips is extracted
    ntips = ntips if isinstance(ntips, int) else ntips.ntips

    # fastest method tested
    return scipy_comb(ntips, subtree_size, exact=True)


def get_num_labeled_trees(ntips: int) -> int:
    r"""Return the number of possible labeled trees for ntips.

    Note
    ----
    The term "labeled trees" has a specific meaning when discussing
    tree space. According to Harding (1971):
    "At each node, the descendant species are partitioned
    into two subtrees. A node at which the subtrees have identical
    unlabeled topologies is balanced; otherwise, it is unbalanced."
    In other words, balanced cherries on a rooted tree contribute
    differently than unbalanced nodes to quantifying labeled trees.

    $$ H_n = \frac{n!(n - 1)!}{2^{n - 1}} $$

    ---------------------------
    | ntips | n labeled trees |
    |-------|-----------------|
    |   3   |        3        |
    |   4   |       18        |
    |   5   |      180        |
    |   6   |     2700        |
    | ...   |      ...        |
    |-------|-----------------|

    See Also
    --------
    ...

    Note
    ----
    Labeled trees

    References
    ----------
    - Edwards (1970)
    - Harding EF (1971) The probabilities of rooted tree-shapes generated by random bifurcation. Adv Appl Prob 3: 44–77.
    - Brown JKM (1994) Probabilities of evolutionary trees. Syst Biol 43: 78–91.
    """
    ntrees = int(
        (factorial(ntips) * factorial(ntips - 1))
        / (2 ** (ntips - 1))
    )
    return ntrees


def get_num_unlabeled_trees(ntips: int) -> int:
    r"""...

    See Harding...

    Note
    ----
    In contrast to labeled trees, all unlabeled trees for ntips are
    *not* equiprobable.
    """



if __name__ == "__main__":

    # get_probability_gene_tree_matches_species_tree(0.001)
    # print(get_num_labeled_trees(6))
    print(get_num_quartets(10))
    # print(get_num_multifurcating_trees(ntips=5, n_internal_nodes=4))



#!/usr/bin/env python

"""Class to construct and return a consensus tree from multiple trees.

Speed
-----
Reducing to only unique topologies costs as much time as just
visiting and computing on them, and would not allow getting dist
values. So this visits all trees.

TODO
----
Support getting mean, etc, of any feature on trees. This is a bit
of work, needs to check all for int,float type. Not done.
"""

from typing import TypeVar, Dict, Optional, Tuple, Iterator
from loguru import logger
import numpy as np
from toytree.core.node import Node
from toytree.core.tree import ToyTree

logger = logger.bind(name="toytree")

MultiTree = TypeVar("MultiTree")


class ConsensusTree:
    """An extended majority rule consensus class.

    Takes a set of input trees and returns a consensus tree. The input
    trees can be rooted or unrooted, and the returned consensus tree
    can be rooted or unrooted, depending on options used.

    Features on the consensus tree. Support values on the consensus
    tree represent the proportion of edges that existed in the set
    of trees.

    Dist values represent the average 'dist' among edges
    that existed in the set of trees, but are only computed if best
    is not supplied as an input tree.

    If ultrametric is set to true then the returned tree is rooted
    on

    Parameters
    ----------
    ultrametric: bool
        If true then node height statistics are calculated and the
        tree will be rooted using the "root_method" argument option.
        The tree can be rooted without being ultrametric, but it
        cannot be ultrametric without being rooted.
    root_method: int
        0: root on midpoint given branch lengths on consensus tree.
        1: root on edge that contains root across majority of input
        trees, using the mean root position on this edge.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(ntips=5, seed=123)
    >>> tree = toytree.rtree.unittree(ntips=5, seed=123)

    Get 20 rooted ultrametric trees w/ same topology and variable dists
    >>> rtrees = [
    >>>    tree.mod.edges_multiplier().mod.edges_slider()
    >>>    for i in range(20)]

    Get 20 unrooted ultrametric trees w/ same topology and variable dists
    >>> utrees = [
    >>>    tree.unroot().mod.edges_multiplier().mod.edges_slider()
    >>>    for i in range(20)]

    Get 20 unrooted non-ultrametric trees w/ same topology and variable dists
    >>> utrees = [
    >>>    tree.unroot().mod.edges_multiplier().mod.edges_slider()
    >>>    for i in range(20)]


    """
    def __init__(
        self,
        mtree: MultiTree,
        best_tree: Optional[ToyTree] = None,
        majority_rule_min: float = 0.0,
        ultrametric: Optional[bool] = None,
        root_method: Optional[bool] = None,
    ):
        # creates an unrooted copy of the original tree
        self.mtree = mtree
        self.best_tree = best_tree
        self.majority_rule_min = majority_rule_min

        if ultrametric is None:
            self.ultrametric = self.mtree.all_tree_tips_aligned()
        else:
            self.ultrametric = ultrametric

    def _iter_unique_trees(self) -> Iterator[Tuple[ToyTree, int]]:
        """Yield unique topologies and their counts from mtree."""
        for tree, count in self.mtree.get_unique_topologies():
            yield tree, count

    def run(self) -> ToyTree:
        """Return a ToyTree with consensus support values."""
        # map clade support values onto a user-supplied input tree
        if self.best_tree is not None:
            return self._map_clades_support_onto_best_tree()

        # build the majrule tree with clade support and other meta data
        ctree = self._get_majority_rule_consensus_tree()

        # if the tree is ultrametric then align tips
        if self.ultrametric:
            # find the most likely root
            ctree.mod.root_on_midpoint(inplace=True)

            # set tips to 0 and root to a temporary high value for now.
            ctree.set_node_data(
                "height",
                {-1: ctree[-2].height + 1} | {i: 0 for i in range(ctree.ntips)},
                inplace=True
            )

            # set internal non-root-associated nodes to their mean height
            ctree.set_node_data(
                "height",
                {i: i.height_mean for i in range(ctree.ntips, ctree.nnodes - 2)},
                inplace=True,
            )

            # prior root node that had its edge split to insert the root
            ctree.set_node_data(
                "height",
                {ctree[-2]: ..., ctree[-1]: ...},
                inplace=True,
            )

            # new root node
            ctree.set_node_data("height", {ctree[-1]: ...}, inplace=True)
        return ctree

    def _map_clades_support_onto_best_tree(self) -> ToyTree:
        """Return the best tree with clade supports from trees.

        This relies on the sorted order of `get_bipartitions`. Finds
        'support' feature for all Nodes in 'best'. Any other requested
        features of the input trees will also be
        """
        # copy best tree and set default to 0
        self.best_tree = self.best_tree.set_node_data("support", default=0)

        # returns ubipartitions (root has no effect)
        best_tree_parts = list(self.best_tree.iter_bipartitions("name", True, False))

        # mirror the support on the root Node's children
        cidxs = [i.idx for i in self.best_tree.treenode.children]
        cmax = max(cidxs)
        cmin = min(cidxs)

        # iterate over all unique tree topologies
        for utree, count in self._iter_unique_trees():

            # iterate over ubipartitions in tree: [[['a'], ['b', 'c']], [...]]
            for nidx, bipart in enumerate(utree.iter_bipartitions("name", True, False)):
                if nidx >= self.best_tree.ntips:
                    try:
                        idx = best_tree_parts.index(bipart)
                        if idx == cmin:
                            self.best_tree[cmax].support += count
                            self.best_tree[cmin].support += count
                        else:
                            self.best_tree[idx].support += count
                    except ValueError:
                        pass

        # divide support by ntrees to get proportion
        ntrees = self.mtree.ntrees
        for node in self.best_tree.traverse():
            node.support /= ntrees

        # root Node doesn't truly have support, except in the sense
        # that an outgroup *was* present and is now trimmed from the
        # tree, while retaining support for this tree versus outside.
        self.best_tree.treenode.support = 1.0
        return self.best_tree

    def _get_majority_rule_consensus_tree(self) -> ToyTree:
        """Return the majrule consensus tree.

        Calculates clade 'support', 'dist', and 'features'.
        This relies on the sorted order of `get_bipartitions`.
        """
        clade_freqs = self._get_all_clade_freqs()
        fclade_freqs = self._get_all_filtered_clades(clade_freqs)
        root = self._build_all_tree(fclade_freqs)
        return ToyTree(root)

    def _build_all_tree(self, fclade_freqs: Dict[Tuple, int]) -> Node:
        """Build majority-rule consensus tree from clades"""

        # root node
        all_tips = frozenset(self.mtree.get_tip_labels())
        # root = Node(name="")
        # cset = fclade_freqs[all_tips]
        # for feat in ['dist_min', 'dist_max', 'dist_median', 'dist_std']:
        #     setattr(root, feat, 0.)
        # for feat in ['height_min', 'height_max', 'height_median', 'height_std']:

        # dict with {tip-set: Node} in order they are added (Py3)
        # sets_to_nodes = {all_tips: root}
        sets_to_nodes = {}

        # visit filtered clades from LARGEST to SMALLEST
        for cset in sorted(fclade_freqs, key=len, reverse=True):

            # skip if already added
            if cset in sets_to_nodes:
                continue

            # create Node to represent this clade
            dists = np.array(fclade_freqs[cset][1])
            heights = np.array(fclade_freqs[cset][2])

            # ...
            node = Node(
                name=str(*cset) if len(cset) == 1 else "",
                support=fclade_freqs[cset][0],
                dist=dists.mean(),
            )

            # ...
            dmin = dists[dists > 0].min() if (dists > 0).sum() else 0
            setattr(node, "dist_min", dmin)
            setattr(node, "dist_max", dists.max())
            setattr(node, "dist_mean", np.mean(dists))
            setattr(node, "dist_median", np.median(dists))
            setattr(node, "dist_std", dists.std())

            hmin = 0 if not (heights > 0).sum() else min(heights[heights > 0])
            setattr(node, "height_min", hmin)
            setattr(node, "height_max", heights.max())
            setattr(node, "height_mean", np.mean(heights))
            setattr(node, "height_median", np.median(heights))
            setattr(node, "height_std", heights.std())

            # visit existing nodes from SMALLEST to LARGEST
            # children iteratively if node is not an descendant.
            for tset in sorted(sets_to_nodes, key=len):
                # is this node (node) is a superset of existing node (tnode)
                if cset.issubset(tset):
                    # get node that exists
                    tnode = sets_to_nodes[tset]
                    tnode._add_child(node)
                    break
            sets_to_nodes[cset] = node
        return sets_to_nodes[all_tips]

    def _get_all_filtered_clades(self, clade_freqs: Dict[Tuple, int]) -> Dict:
        """Remove conflicting clades and those < majority_rule_min"""
        # keep track of kept clades in order they are kept.
        keep_dict = {}  # {frozenset : float}

        # visit clades in sorted order and drop conflicts or low support
        for clade, (freq, dist, height) in clade_freqs.items():

            # clade is below threshold, discard.
            if freq < self.majority_rule_min:
                # logger.debug(f"clade {clade}:{freq} below min threshold")
                continue

            # clade conflicts with others, discard. If any tips in this
            # clade also occur in an existing clade, then this must be
            # a subset or superset of that clade. Example:
            # ('a', 'b', 'c') conflicts with ('a', 'b', 'd')
            # ('a', 'b', 'c') does not conflict with ('a', 'b', 'c', 'd')
            # ('a', 'b', 'c') does not conflict with ('a', 'b')
            cset = frozenset(clade)

            # check until conflict, or not more sets to compare
            conflict = False
            for pset in keep_dict:
                if conflict:
                    break
                if cset.isdisjoint(pset):
                    continue
                if cset.issubset(pset):
                    continue
                if cset.issuperset(pset):
                    continue
                conflict = True
                # logger.debug(f"clade {clade} conflicts.")

            # passed filters, keep it.
            if not conflict:
                keep_dict[cset] = (freq, dist, height)
        return keep_dict

    def _get_all_clade_freqs(self) -> Dict:
        """Return a dict of {clades: features} where 'support' feature
        is frequency of occurrence across treelist.

        Additional features are stored in a list in order following
        'support' and 'dist'. Clades are represented by the smaller
        of side of a bipartition (or lowest name str if same size).

        The returned dict is sorted by high->low support values.
        {clade2: [support, dist, ...], clade2: [support, ...], ...}
        """
        # keep track of all observed clades
        clades = {}
        ntrees = self.mtree.ntrees
        increment = 1 / ntrees

        # iterate over all unique tree topologies
        for utree in self.mtree:

            # iterate over ubipartitions in tree: [[['a'], ['b', 'c']], [...]]
            for nidx, bipart in enumerate(utree.iter_bipartitions("name", True, False)):
                # skip tip-only Nodes
                node = utree[nidx]
                # store the smaller half
                clade = tuple(bipart[0])

                # if a root child then store the edge length in unrooted form
                # store the root position on the branch in the case that it
                # is again inferred as the root.
                if node.up:
                    if node.up.is_root():
                        children = node.up.children
                        if len(children) == 2:
                            dist = sum(i.dist for i in children)
                            logger.debug([node, node.dist, dist])
                        else:
                            dist = node.dist
                    else:
                        dist = node.dist
                else:
                    dist = node.dist

                # store dist and height
                if clade in clades:
                    clades[clade][0] += increment
                    clades[clade][1].append(dist)
                    clades[clade][2].append(node.height)
                else:
                    clades[clade] = [increment, [node.dist], [node.height]]

        # return in sorted order and w/ counts as proportions
        sort_clades = sorted(
            clades,
            key=lambda x: clades[x][0], reverse=True
        )

        # add the full (all samples) clade to get stats for it.
        all_tips = tuple(self.mtree[0].get_tip_labels())
        sort_clades = [all_tips] + sort_clades
        clades[all_tips] = (
            1.0,
            [i.treenode.dist for i in self.mtree],
            [i.treenode.height for i in self.mtree],
        )
        return {i: clades[i] for i in sort_clades}


if __name__ == "__main__":

    import toytree

    # these trees should all have full support
    tres = [toytree.rtree.unittree(10, seed=123) for i in range(5)]
    mtre = toytree.mtree(tres)
    ctre = ConsensusTree(mtre).run()
    # ctre = ctre.root("r8", "r9")
    # ctre._draw_browser(node_sizes=20, node_labels="support")
    print(ctre.write(features=ctre.features))

    # ctre = ConsensusTree(mtre, best_tree=mtre[0]).run()
    # ctre = ctre.root('r8', 'r9')
    # ctre._draw_browser(node_sizes=20, node_labels="support")

    # these trees should all have diff support
    # trees = [toytree.rtree.unittree(10, ) for i in range(20)]
    # mtree = toytree.mtree(trees)
    # ctree = ConsensusTree(mtree, best_tree=mtree[0]).run()
    # ctree._draw_browser(node_sizes=20, node_labels="support")
    # ctree.root('r8', 'r9', inplace=True)
    # ctree._draw_browser(node_sizes=20, node_labels="support")



#!/usr/bin/env python

"""Function to get data from all Nodes in a tree.

Data is returned as a pandas Dataframe and any value can be imputed
for missing values for Nodes that do not have assigned data for a
feature.
"""

from typing import Union, Sequence, Any, TypeVar, Optional
from loguru import logger
import pandas as pd
import numpy as np
from toytree import ToyTree, Node
from toytree.core.apis import add_toytree_method
# from toytree.utils import ToytreeError

Query = TypeVar("Query", int, str, Node)
logger = logger.bind(name="toytree")


@add_toytree_method(ToyTree)
def get_node_data(
    tree: ToyTree,
    feature: Union[str, Sequence[str], None] = None,
    missing: Union[Any, Sequence[Any], None] = None,
) -> Union[pd.DataFrame, pd.Series]:
    """Return a pandas Series or DataFrame with values for one or
    more selected features in the tree.

    This function is convenient for accessing data in tabular
    format, but is slower than accessing data directly from Nodes,
    for example during a traversal, because it spends time checking
    for Nodes with missing values and type-checking missing values.
    See Documentation for accessing data from Nodes.

    Setting complex objects to Node data, such as lists or sets,
    rather than float, int, or str, should generally work fine,
    but beware that this function will not attempt to automatically
    check or fill missing values for these data.

    See Also: `get_feature_dict`, `set_node_data`, `get_tip_data`.

    Parameters
    ----------
    feature: str, Sequence[str], or None
        One or more features of Nodes to get data for. Features
        include the default Node features (idx, name, height, dist,
        support) as well as any attribute that has been set to
        a Node (except attrs with names that start with an '_'.)
    missing: Any, Sequence[Any], or None
        A value to use for missing data (Nodes that do not have
        the feature). Default arg is None which will automatically
        set missing data to `np.nan`. Example: you can set the
        missing values to a default value like 0 for an int feature
        or by entering 0, or you can enter a list of values to
        set default missing for all features.

    Returns
    -------
    pd.DataFrame or pd.Series
        If a single feature is selected then a pd.Series will be
        returned with tip node 'idx' attributes as the index.
        If multiple features are selected (or None, which selects
        all features) then a pd.DataFrame is returned with tip
        node 'idx' attributes as the index and feature names as
        the column labels.

    Examples
    --------
    Add a new feature to some nodes and fetch data for all nodes.
    >>> tree = toytree.rtree.unittree(10)
    >>> tree = tree.set_node_data("trait1", {0: "A", 1: "B"})
    >>> tree = tree.set_node_data("trait2", {2: 3.5, 3: 5.0})
    >>> data1 = tree.get_tip_data(feature="trait1", missing="C")
    >>> data2 = tree.get_tip_data(feature="trait2")
    """
    # Note: tried storing missing as pd.NA which allows not having to
    # convert the dtype of other values from e.g., int to float. BUT, if
    # <NA> gets converted to a string for plottig it causes total havoc
    # on the HTML. And anyways it looks weird to have a mix of NaN
    # and <NA> in the data table. So I chose np.nan instead of pd.NA.
    # Also pd.NA is very slow compared to np.nan.

    # TODO: AVOID FORMATTING FOR COMPLEX FEATURE TYPES (E.G., DICT, SET, ETC).

    # select one or more features to fetch values for
    if feature is None:
        features = tree.features
    elif isinstance(feature, (list, tuple)):
        features = feature
    else:
        features = [feature]

    # create a list of missing values for subs
    if missing is None:
        missing = [np.nan] * len(features)
    elif isinstance(missing, (list, tuple)):
        assert len(missing) == len(features), (
            "when entering multiple missing values it must be the same "
            "length as the number of features")
    else:
        missing = [missing] * len(features)

    # check for bad user features
    for feat in features:
        if feat not in tree.features:
            raise ValueError(f"feature '{feature}' not in tree.features.")

    # store as ordered lists, and let pd.Series convert to dtype
    data = {}
    for fidx, feat in enumerate(features):

        # fill ordered list with Node value or missing value
        ofeat = []
        miss = missing[fidx]

        # if miss is None then find auto-filling type. This is
        # quite slow (milliseconds) making this not the recommended
        # method for quick data fetching, as explained in docs.
        # if miss is None:
        #     values = [getattr(self[nidx], feat, None) for nidx in range(self.nnodes)]
        #     types = [type(i) for i in values if i is not None]

        # get value or set to missing
        for nidx in range(tree.nnodes):
            value = getattr(tree[nidx], feat, miss)

            # if the actual value is nan then replace with miss
            try:
                if np.isnan(value):
                    value = miss
            except (TypeError, ValueError):
                pass
            ofeat.append(value)

        # allow pandas to infer dtype
        series = pd.Series(ofeat)

        # store Series to a dict
        data[feat] = series

    # if a single feature was selected return as a Series else DataFrame
    if len(features) == 1:
        return series
    return pd.DataFrame(data)


@add_toytree_method(ToyTree)
def get_tip_data(
    tree: ToyTree,
    feature: Union[str, Sequence[str], None] = None,
    missing: Optional[Any] = None,
) -> pd.DataFrame:
    """Return a DataFrame with values for one or more selected
    features from every leaf node in the tree.

    Parameters
    ----------
    feature: str, Iterable[str], or None
        One or more features of Nodes to get data for.
    missing: Any
        A value to use for missing data (nodes that do not have
        the feature). Default arg is None which will automatically
        select a missing value based on the data type. Example:
        "" for str type, np.nan for numeric or complex types.
        Any value can be entered here to replace missing data.

    Returns
    -------
    data: pd.DataFrame or pd.Series
        If a single feature is selected then a pd.Series will be
        returned with tip node 'idx' attributes as the index.
        If multiple features are selected (or None, which selects
        all features) then a pd.DataFrame is returned with tip
        node 'idx' attributes as the index and feature names as
        the column labels.

    Examples
    --------
    Add a new feature to some nodes and fetch data for all nodes.
    >>> tree = toytree.rtree.unittree(10)
    >>> tree = tree.set_node_data("trait1", {0: "A", 1: "B"})
    >>> tree = tree.set_node_data("trait2", {2: 3.5, 3: 5.0})
    >>> data1 = tree.get_tip_data(feature="trait1", missing="C")
    >>> data2 = tree.get_tip_data(feature="trait2")

    See Also
    --------
    get_feature_dict
        Get a dict mapping any node feature to another.
    set_node_data
        Set a feature value to one or more Nodes in a ToyTree.

    Note
    ----
    This function is convenient for accessing data in tabular
    format, but is slightly slower than accessing data directly
    from Nodes because it spends time type-checking missing data.
    """
    return tree.get_node_data(feature, missing).iloc[:tree.ntips]


if __name__ == "__main__":

    import toytree
    toytree.set_log_level("INFO")

    tree = toytree.rtree.unittree(ntips=10)

    # tree = tree.set_node_data("x", {11: "red"}, default="blue", inherit=True)
    # print(get_node_data(tree, 'x'))

    tree = tree.set_node_data("x", {11: "red"}, inherit=True)
    print(get_node_data(tree, 'x', missing='red'))



#!/usr/bin/env python

"""toytree color subpackage.

This is a simple package for accessing convenience functions for 
selecting color palettes. See `toyplot.color` for the full color
module.

Examples
--------
>>> tree.draw(node_colors=toytree.colors.COLOR1[0])
>>> tree.draw(node_colors=toytree.colors.COLOR1)

"""

from toytree.color.src.utils import COLORS1, COLORS2, color_cycler
from toytree.color.src.toycolor import ToyColor
from toytree.color.src.colorkit import ColorType
from toytree.color.src.concat import concat_style_fix_color
# from toytree.color.src.color_mapper import get_color_mapped_feature



#!/usr/bin/env python

"""Fit a discrete Markov model to data using Maximum Likelihood.

Model fit estimates model parameters and ancestral states.

Example
-------
>>> toytree.pcm.fit_discrete_markov_data(
        nstates=3, model="ER", 
        rates=None,               # will be estimated
        state_frequencies=None,   # will be estimated
    )
"""

from typing import Optional, Union, List, TypeVar
from dataclasses import dataclass, field
import pandas as pd
import numpy as np
from scipy.optimize import minimize
from scipy.linalg import expm
from toytree.utils import ToytreeError
from toytree.pcm.src.discrete_markov_model_sim import MarkovModel


ToyTree = TypeVar("ToyTree")


@dataclass
class FitMarkovModelResult:
    """Result object of a Markov model fit by Maximum Likelihood."""
    nstates: int
    model: str
    transition_matrix: np.ndarray
    state_frequencies: np.ndarray
    data: pd.DataFrame = None
    log_likelihood: float = None
    fixed_rates: np.ndarray = None
    nparams: int = None


@dataclass
class FitMarkovModelBase:
    """Fit a Markov model to discrete data using Maximum Likelihood.

    This uses Felsenstein's pruning algorith to recursively traverse
    a tree to compute conditional likelihoods of each state at each
    Node, and returns the total likelihood as the sum of 
    log-likelihoods of the states at the root.

    Parameters
    ----------
    tree: ToyTree
        A ToyTree instance (usually ultrametric).
    data: pandas.Series or pandas.DataFrame
        If the index is numeric it is interpreted as node idx labels,
        if not, then as string tip labels. Column names are not used,
        and if multiple columns present they are treated as replicate 
        observations of the same trait. Data is treated as categorical
        (0-nstates). 
    nstates: int
        The number of discrete states in the Markov model.
    model: str
        The Markov model type ("ER", "SYM", "ARD") will determine the
        number of parameters to be inferred. The fixed_rates arg can
        be used to fix parameters, which will then not be inferred.
    fixed_rates: Optional[np.ndarray]
        An array of the dimension (nstates,nstates) with fixed rate
        parameters (not to be estimated) in some or all off-diagonal
        cells, or np.nan for parameters that should be estimated. 
        Fixing parameters decrease the number that needs to be 
        estimated. Fixed rates must match allowed values given the 
        model type (e.g., ER must fix all to a single rate b/c only
        on rate is allowed in this model, and SYM model rates must be
        symmetric). Values on the diagonal are ignored.
    fixed_state_frequencies: Optional[np.ndarray]
        An array of dimension nstates with fixed values for the state
        frequencies... how does this relate to root prior...?

    Examples
    --------
    >>> #Fit a 2-state character observed for all tips in the tree.
    >>> tree = ...
    >>> data = tree.pcm.simulate_discrete_data(3, "ER", tips_only=True)
    >>> fit = tree.pcm.fit_discrete_data(data, 3, "ER")
    >>>
    >>> #Fit a 3-state SYM model with observations for some internal nodes.
    >>> data = tree.pcm.simulate_discrete_data(3, "ER", tips_only=True)
    >>> fit = tree.pcm.fit_discrete_data(data, 3, "ER")
    >>>
    >>> #Fit a 2-state ARD model with some rates fixed:
    >>> data = tree.pcm.simulate_discrete_data(2, "ARD",
    >>>    relative_rates=[[0, 2], [1, 0]])
    >>> fixed = np.array([[0, 2],[np.nan, 0]])
    >>> fit = tree.pcm.fit_discrete_data(data, 2, "ARD", fixed_rates=fixed)
    >>>
    >>> #Fit a GTR model to DNA-like 4-state data
    >>> true_rates = 1e-8 * np.array([
    >>>     [0, 1, 2, 2],
    >>>     [1, 0, 2, 2],
    >>>     [2, 2, 0, 1],
    >>>     [2, 2, 1, 0],
    >>> ])
    >>> true_freqs = [0.2,0.3,0.2,0.3]
    >>> data = tree.pcm.simulate_discrete_data(4, "SYM",
    >>>    relative_rates=true_rates, 
    >>>    state_frequencies=true_freqs,
    >>>    rate=1e-8, 
    >>>    nreplicates=500, 
    >>>    tips_only=True,
    >>> )
    >>> fit = tree.pcm.fit_discrete_data(data, 4, "SYM")

    References
    ----------
    - Yang.
    - Paradis et al. (2004).
    - Harmon book.
    """
    tree: ToyTree
    data: Union[pd.Series, pd.DataFrame]
    nstates: int
    model: str

    root_prior: Optional[np.ndarray]=None   
    fixed_rates: Optional[np.ndarray]=None
    fixed_state_frequencies: Optional[np.ndarray]=None

    nparams_fixed: int=field(init=False)
    nparams_free: int=field(init=False)
    rate_matrix: np.ndarray=field(init=False)
    state_frequencies: np.ndarray=field(init=False)

    def __post_init__(self):
        self.model = self.model.upper()
        assert self.model in ("ER", "SYM", "ARD"), (
            "model must be one of 'ER', 'SYM', 'ARD'")

        # Set parameters to be inferred as np.nan, and fixed params
        # to the fixed value, and fill nparams_ counters.
        self._check_freqs()
        self._check_rates_and_nparams()

        # ...
        # self._check_data_with_tree()

        # self._check_rates()
        # self.qmatrix = self._get_qmatrix()

    def _check_freqs(self):
        """Check stationary frequencies given model and nstates.

        Entered values are checked for correct size and that they 
        sum to one. If no values are entered then a uniform frequency
        of the correct size is sampled to represent the best starting
        guess for root state.

        Sets values for array of shape to nstates - 1, e.g., 
        infer states for k=3: [np.nan, np.nan]
        one fixed state for k=3: [0.5, np.nan]
        """
        freqs = self.state_frequencies
        if freqs is None:
            freqs = np.repeat(1.0 / self.nstates, self.nstates)
        else:
            freqs = np.array(freqs)
            if self.model in ("SYM", "ARD"):
                assert freqs.size == self.nstates, (
                    f"states_frequencies should be len={self.nstates}.")
            else:
                fixed = np.repeat(1.0 / self.nstates, self.nstates)
                assert np.allclose(freqs, fixed), (
                    f"ER model with nstates={self.nstates} has fixed state "
                    f"frequences: {fixed}. See SYM or ARD models.")
                freqs = fixed
        assert np.allclose(sum(freqs), 1.0), (
            f"state_frequencies must sum to 1. You entered: {freqs}")
        self.state_frequencies = freqs

    def _check_rates_and_nparams(self):
        """Get nparams based on model, nstates, and fixed params.

        Checks that fixed params is valid given the other args. For
        example, you can fix all parameters in which case no ML
        optimization occurs and the loglik is simply calculated.

        Number of parameters (k=nstates): The "ER" model always has 1 
        parameter; the "SYM" model has ((k * (k-1)) / 2) + (k - 1) 
        parameters; the "ARD" is (k * (k-1)) + (k - 1). The effect of
        fixing a parameter is -2k for SYM, or -k for others.
        """
        rates = np.ones((self.nstates, self.nstates))
        assert rates.shape == self.fixed_rates, (
            "fixed_rates array must be dimension (nstates, nstates)")
        np.fill_diagonal(rates, 0)
        np.fill_diagonal(self.fixed_rates, 0)

        # fixed rates must by symmetric for ER and SYM
        if self.model in ("ER", "SYM"):
            assert np.allclose(self.fixed_rates, self.fixed_rates.T, rtol=1e-5, atol=1e-8), (
                "rates must be symmetric in ER and SYM models. See ARD model.")

        k = self.nstates
        mask = np.invert(np.eye(self.nstates, dtype=bool))
        fixed = self.fixed_rates[mask]
        if self.model == "ER":
            self.nparams_fixed = fixed[~np.isnan(fixed)].size
            self.nparams_free = 1 - self.nparams_fixed
        elif self.model == "SYM":
            self.nparams_free = (k * (k - 1)) / 2
            self.nparams_fixed = 2 * fixed[~np.isnan(fixed)].size
            self.nparams_free = 1 - self.nparams_fixed
        else:
            self.nparams_fixed = fixed[~np.isnan(fixed)].size
            self.nparams_free = (k * (k - 1)) - self.nparams_fixed

    def _check_data_with_tree(self):
        """TODO... """


class FitMarkovModel(FitMarkovModelBase):
    """...

    Examples
    --------
    >>> tree = toytree.rtree.unittree(10, seed=123)
    >>> data = toytree.pcm.simulate_discrete_data(tree, 3, "ER")
    >>>
    >>> # Fit an ER model to the data
    >>> fit = toytree.pcm.fit_discrete_markov_model(tree, data, "ER")
    >>> print(fit.log_likelihood, fit.data)
    """
    def __init__(
        self,
        log_likelihood: float,
        data: pd.DataFrame,
        ):
        pass

    def _check_data_with_tree(self):
        """Data index must be a set length and type.

        Data will be converted to a DataFrame with index=range(nnodes)
        and discrete data as floats 0-nstates or np.nan for missing.
        """
        assert isinstance(self.data, pd.DataFrame), "data must be a DataFrame"
        # if data index is strings then convert to node idxs

        assert set(self.data.index) == set(self.tree.get_tip_labels()), (
            "the index of data must have the same tip names as the input tree")



class DiscreteMarkovModelFit:
    """Fit parameters of an n-state discrete Markov model using 
    maximum likelihood.

    Uses scipy to estimate transition rate parameters under one of
    several supported models ("ER", "SYM", ARD").

    Parameters
    ----------
    tree: ToyTree
        A toytree object with ultrametric branch lengths.
    data: pd.DataFrame
        A DataFrame with tip names as the index, and categorical data 
        values (e.g., int, str, etc.) assigned to every tip in the tree.
        The column names are not used. All columns in the DataFrame are
        analyzed as replicate observations.
    model: str
        ...

    Examples
    --------
    Simulate discrete data on a tree.
    >>> tree = toytree.rtree.unittree(ntips=10, treeheight=10)
    >>> data = tree.pcm.simulate_discrete_data(tree, 3, "ER", rate=0.5)
    
    Fit Markov model to discrete data.
    >>> fit = tree.pcm.fit_discrete_markov_model(tree, data, 3, "ER")
    >>> print(fit.log_likelihood, fit.data)

    See also
    --------
    :func:`toytree.pcm.simulate_discrete_markov_data`

    References
    ----------
    https://github.com/rcohen/hogtie
    """
    def __init__(
        self, 
        tree: 'toytree.ToyTree',
        data: pd.DataFrame,
        model: str,
        prior: Optional[np.ndarray]=None,
        ):

        self.tree = tree    # get's replaced by a copy 
        self.model = model

        # get unique data and ensure tip names aligned
        if isinstance(data, pd.Series):
            data = pd.DataFrame(data)
        assert isinstance(data, pd.DataFrame), "data must be a DataFrame"
        assert set(data.index) == set(tree.get_tip_labels()), (
            "the index of data must have the same tip names as the input tree")
        self.data = data.copy() # gets converted to ints

        # get empty Q matrix ...
        self.ints_to_states = dict(enumerate(np.unique(data)))
        self.states_to_ints = {j:i for i, j in self.ints_to_states.items()}
        self.nstates = len(self.states_to_ints)
        self.prior = (
            prior if prior is not None 
            else np.repeat(1 / self.nstates, self.nstates)
        )
        assert self.prior.size == self.nstates


        # organize data as ints to w/ only uniques
        self.unique = None
        self.inverse = None
        self.get_unique_data()

        # the params to infer depends on the model type and size
        self.qmatrix = np.zeros((self.nstates, self.nstates))
        self.params = {}
        self.qidx = None
        if isinstance(self.model, str):
            self.model = self.model.upper()
            self.set_qmatrix_auto()
        else:
            self.set_qmatrix_custom()

        # prepare data
        self.set_node_arrays_to_tree()


    def get_unique_data(self):
        """
        Record only unique observed tip patterns and the inverse info
        to expand back to the full later.
        """
        for state in self.states_to_ints:
            self.data[self.data == state] = self.states_to_ints[state]
        self.unique, self.inverse = np.unique(
            self.data,
            return_inverse=True,
            axis=1,
        )
        #print(f'unique array shape: {self.unique.shape}')


    def set_qmatrix_custom(self):
        """
        TODO: Parse the custom qmatrix from user and check:
          - todo: cannot enter mixed type w/ numpy tho...
          - Values on the diagonal cannot be set (are overwritten)
          - Values as floats are fixed at their value (not inferred)
          - Values as ints on any off-diagonal are inferred.
        """
        raise NotImplementedError("TODO")
        # try to coerce into an array if not a string
        starting = 1 / self.tree.treenode.height
        self.qidx = np.array(self.model).astype(int)
        self.params = {i: starting for i in np.unique(self.qidx) if i}


    def set_qmatrix_auto(self):
        """
        Get the number of parameters to estimate, and a matrix (qidx)
        with those parameters. 

               ER          SYM          ARD        CUSTOM
           [0 1 1 1]    [0 1 2 3]    [0 1 2 3]      ...
           [1 0 1 1]    [1 0 1 2]    [7 0 4 5]
           [1 1 0 1]    [2 1 0 1]    [8 9 0 6]
           [1 1 1 0]    [3 2 1 0]    [10 11 12 0]

            [0 1 1]      [0 1 2]      [0 1 2]
            [1 0 1]      [1 0 3]      [4 0 3]
            [1 1 0]      [2 3 0]      [5 6 0]

             [0 1]        [0 1]        [0 1]
             [1 0]        [2 0]        [2 0]
        """
        starting = 1 / self.tree.treenode.height        
        self.qidx = np.zeros(self.qmatrix.shape, dtype=int)

        # assign everything to the same one rate (ER)
        if self.model.upper() == "ER":
            self.params[1] = starting
            self.qidx[:] = 1
            self.qidx[np.diag_indices_from(self.qidx)] = 0
            return

        # ARD & SIM: assign a parameter to every cell in upper triangle
        pidx = 1
        indices = np.triu_indices(self.qidx.shape[0], k=1)
        for cidx, cidy in zip(*indices):
            self.params[pidx] = 1 / self.tree.treenode.height
            self.qidx[cidx, cidy] = pidx
            pidx += 1

        # assign a parameter to every cell in upper triangle            
        indices = np.tril_indices(self.qidx.shape[0], k=-1)
        for cidx, cidy in zip(*indices):
            if self.model == "ARD":
                self.params[pidx] = starting
                self.qidx[cidx, cidy] = pidx
                pidx += 1
            elif self.model == "SYM":
                self.qidx[cidx, cidy] = self.qidx[cidy, cidx]
            else: 
                raise ToytreeError(f"model type {self.model} not recognized")


    def set_node_arrays_to_tree(self):
        """
        Sets a numpy array as a feature to every node that will store
        the likelihood of each state in each replicate (nuniq, nstates)       

        3-state model: 
                 [0.5, 0.5, 0]
                 [0, 0.5, 0.5]
                      ...
                 /           \
            [1, 1, 0]     [1, 1, 0]
            [0, 0, 1]     [0, 1, 0]
               ...           ...
        """
        value = np.zeros((self.unique.shape[1], self.qidx.shape[0])),
        self.tree = self.tree.set_node_data("likelihood", default=value)
        for nidx in range(self.tree.ntips):
            node = self.tree[nidx]

            # get column index of this tip
            sidx = self.data.index.tolist().index(node.name)

            # get column from unique matching this tip index
            udata = self.unique[sidx, :]
            for istate in self.ints_to_states:
                node.likelihood[:, istate] = (udata == istate).astype(float)


    def node_conditional_likelihoods(self, node):
        """
        Returns the conditional likelihood at a single node given the
        likelihood's of data at its child nodes.
        """
        # get transition probabilities over each branch length
        probmats = {
            0: expm(self.qmatrix * node.children[0].dist),
            1: expm(self.qmatrix * node.children[1].dist),
        }

        nstates = node.likelihood.shape[1]
        anclik = []
        for ancstate in range(nstates):
            # likelihood of child0 observation if anc==ancstate
            prob = (probmats[0][ancstate, :] * node.children[0].likelihood).sum(axis=1)

            prob = (probmats[1][:, 0] * node.children[1].likelihood).sum(axis=1)            

            # TODO...
            # anclik[ancstate] = x

            for endstate in range(nstates):
                for child in range(2):
                    prob = probmats[child]
                    prob[endstate, 0] * node.children[0].likelihood[:, ]

        anc_liks = []
        for childidx in range(2):
            prob = probmats[childidx]
            for sidx in range():

                # prob child==sidx if anc==sidx
                prob[sidx, sidx]

                # get into a notebook for this...

        # likelihood that child 0 observation occurs if anc==0
        print(prob_child0, '0000')
        child0_is0 = (
            prob_child0[0, 0] * node.children[0].likelihood[:, 0] + 
            prob_child0[0, 1] * node.children[0].likelihood[:, 1]
        )

        # likelihood that child 1 observation occurs if anc==0
        child1_is0 = (
            prob_child1[0, 0] * node.children[1].likelihood[:, 0] + 
            prob_child1[0, 1] * node.children[1].likelihood[:, 1]
        )
        anc_lik_0 = child0_is0 * child1_is0

        # likelihood that child 0 observation occurs if anc==1
        child0_is1 = (
            prob_child0[1, 0] * node.children[0].likelihood[:, 0] + 
            prob_child0[1, 1] * node.children[0].likelihood[:, 1]
        )
        child1_is1 = (
            prob_child1[1, 0] * node.children[1].likelihood[:, 0] + 
            prob_child1[1, 1] * node.children[1].likelihood[:, 1]
        )
        anc_lik_1 = child0_is1 * child1_is1

        # set estimated conditional likelihood on this node
        node.likelihood = np.column_stack([anc_lik_0, anc_lik_1])        


    def pruning_algorithm(self):
        """
        Traverse tree from tips to root calculating conditional 
        likelihood at each internal node on the way, and compute final
        conditional likelihood at root based on priors for root state.
        """
        # traverse tree in order **from tips to root**
        # to get conditional likelihood estimate at root.
        for node in self.tree.treenode.traverse("postorder"):
            if not node.is_leaf():               
                self.node_conditional_likelihoods(node)

        # multiply root prior times the conditional likelihood at root
        root = self.tree.treenode
        print(self.prior)
        print(root.likelihood)
        lik = self.prior * root.likelihood
        print(lik)
        lik = lik.sum()
        print(lik)
            # (1. - self.prior_root_is_1) * root.likelihood[:, 0] + 
            # self.prior_root_is_1 * root.likelihood[:, 1]
        # )
        return lik[self.inverse]


    def optimize(self):
        """
        Use maximum likelihood optimization to find the optimal 
        parameters (from .params) to fit the data.

        TODO: max bounds could be set based on tree height. For smaller
        tree heights (e.g., 1) the max should likely be higher. If the 
        estimated parameters is at the max bound we should report a 
        logger.warning(message).
        """  
        estimate = minimize(
            fun=optim_func,
            x0=np.array([self.params[i] for i in self.params]),
            args=(self,),
            method='L-BFGS-B',
            bounds=[(1e-12, 500) for i in self.params],
        )
        self.model_fit = {
            "params": estimate.x,
            "negLogLik": estimate.fun,
            "convergence": estimate.success,
        }
        print(self.model_fit)
        print(self.qmatrix)
        print(-np.log(self.pruning_algorithm()))



def optim_func(params, model):
    """Function to optimize for discrete Markov Model. 

    Takes an iterable as the first argument 
    containing the parameters to be estimated (alpha, beta), and the
    BinaryStateModel class instance as the second argument.
    """
    # update qmatrix with input params
    for pidx, param in enumerate(params):
        idxs = np.where(model.qidx == pidx + 1)
        model.qmatrix[idxs] = params[param]
    for idx in range(model.qmatrix.shape[0]):
        rowsum = model.qmatrix[idx].sum()
        model.qmatrix[idx, idx] = -(rowsum - model.qmatrix[idx, idx])        

    # get likelihood of the data given the model parameters
    liks = model.pruning_algorithm()
    return -np.log(liks).sum()



if __name__ == "__main__":

    import toytree

    TREE = toytree.rtree.unittree(5, treeheight=1)
    print(toytree.pcm.get_markov_model(3, "ER", rate=1).transition_matrix)
    DATA = toytree.pcm.simulate_discrete_data(TREE, 3, "ER", rate=1)
    print(DATA)

    fit = FitMarkovModelBase(TREE, DATA, 3, "ER")
    print(fit)


    # print(TREE.ntips, len(TREE.get_tip_labels()))
    # DATA = DATA[:TREE.ntips]
    # DATA.index = TREE.get_tip_labels()
    # print(DATA)

    # MOD = DiscreteMarkovModelFit(TREE, data=DATA, model="SYM")
    # print(MOD.qidx)
    # print(MOD.params)
    # print(MOD.unique)
    # print(MOD.tree[0].likelihood)
    # MOD.optimize()
    # # print(MOD.prior)
